"""
-------------------------------------------------------------------------------

Tools to compute spectra from a transfer function, and various helpers, too.

-------------------------------------------------------------------------------
"""

import numpy as np
from enum import Enum
import scipy.integrate as integrate


# noinspection SpellCheckingInspection,PyPep8Naming
def calculate_gain_spectrum(transfer_function_panel: np.ndarray) -> np.ndarray:
    """
    The gain spectrum g(w) is

        g(w) = |H(jw)|,

    where w = 2 pi f. The input and returned np arrays read

        [f, H(j 2 pi f)]
    and
        [f,  g(f)].

    Parameters
    ----------
    transfer_function_panel: np.ndarray
        Transfer function panel with cols [f, H(j 2 pi f)].

    Returns
    -------
    panel: np.ndarray
        Gain spectrum panel with cols [f, g(2 pi f)].
    """

    # take the absolute value of the complex values
    panel = make_conformant_float_panel_and_copy_axis_0(transfer_function_panel)
    panel[:, 1] = abs(transfer_function_panel[:, 1])

    # return
    return panel


# noinspection SpellCheckingInspection,PyPep8Naming
def calculate_group_delay_spectrum(
    phase_spectrum_panel: np.ndarray,
) -> np.ndarray:
    """
    Returns the group-delay spectrum from an input phase spectrum.
    The group-delay spectrum is defined with respect to the phase
    spectrum phase(w) as

        grp_delay(w) = - d phase(w) / dw .

    The derivative is taken numerically as a first difference over
    two grid steps.

    Parameters
    ----------
    phase_spectrum_panel: np.ndarray
        Transfer function panel with cols [f, angle_H(j 2 pi f)].

    Returns
    -------
    panel: np.ndarray
        Group-delay spectrum panel with cols [f, group-delay(2 pi f)].
    """

    # make conformant group-delay panel
    panel = phase_spectrum_panel.copy()

    # infer df, calc dw
    df = min(np.diff(phase_spectrum_panel[:, 0]))

    # take 1st difference two steps apart (for O(2) error)
    # mark first and last entries as np.nan since those values are
    panel[1:-1, 1] = -(panel[2:, 1] - panel[:-2, 1]) / (2.0 * 2.0 * np.pi * df)
    panel[0, 1] = np.nan
    panel[-1, 1] = np.nan

    # return
    return panel


# noinspection SpellCheckingInspection,PyPep8Naming
def calculate_spectra(H: np.ndarray, phase_H: np.ndarray) -> np.ndarray:
    """Calculates gain, phase and group delay from H and angle-H.

    Parameters
    ----------
    H: np.ndarray
        Transfer function values in frequency.
    phase_H
        Phase of transfer function values in frequency.

    Returns
    -------
    spectra: np.ndarray
        Spectra: (n, 4) with cols [f, gain(f), phase(f), group-delay(f)]
    """

    # alloc ndarray [n, 4] with cols [f, gain(f), phase(f), grp-delay(f)]
    spectra = np.zeros((H.shape[0], 4))

    # call up the column enum
    cols = SpectraColumns

    # frequency axis
    spectra[:, cols.FREQ.value] = np.real(H[:, 0].copy())

    # spectra
    spectra[:, cols.GAIN.value] = calculate_gain_spectrum(H)[:, 1]

    spectra[:, cols.PHASE.value] = phase_H[:, 1].copy()

    spectra[:, cols.GROUPDELAY.value] = calculate_group_delay_spectrum(phase_H)[
        :, 1
    ]

    # return
    return spectra


# noinspection SpellCheckingInspection,PyPep8Naming
def perturb_dc_frequency(
    transfer_function_panel: np.ndarray, scale=1.0e-6
) -> np.ndarray:
    """
    In the case that there is a zero value for time or frequency in
    the panel, this function replaces that value with a perturbated value.
    This function is useful when computing transfer functions that have
    poor behavior around f = 0 (eg an mbox level), or for phase
    spectra where the magnitude of the transfer function is zero at DC.

    Parameters
    ----------
    transfer_function_panel: np.ndarray
        Transfer function panel with cols [f, H(j 2 pi f)].
    scale: float
        Scale factor applied to the perturbation step.

    Returns
    -------
    transfer_function_panel: np.ndarray
        Transfer function panel with any 0 values in col 0 perturbed.
    """

    # pull the freq axis
    f = np.real(transfer_function_panel[:, 0])
    df = min(np.diff(f))

    # see if there is a replacement to make
    replace_indices = np.where(f == 0.0)[0]

    # replace as necessary
    if len(replace_indices) > 0:

        transfer_function_panel[replace_indices, 0] = scale * df

    # return
    return transfer_function_panel


# noinspection SpellCheckingInspection,PyPep8Naming
def make_conformant_float_panel_and_copy_axis_0(transfer_function_panel):
    """."""

    # alloc a zeros panel, copy over the freq axis
    panel = np.zeros_like(transfer_function_panel, dtype=float)
    panel[:, 0] = np.real(transfer_function_panel[:, 0])

    return panel


def wrap_phase(phase: np.ndarray) -> np.ndarray:
    """Wraps input phase to the domain [-pi, pi].

    This is an imperfect function. The map +pi -> +pi works, but the
    map for -pi is -pi -> +pi. The meaning of this function wrap()
    is itself ambiguous when applied to the entire real number line,
    rather than non-negative numbers.

    Parameters
    ----------
    phase: np.ndarray
        Input phase array.

    Returns
    -------
    np.ndarray
        Output phase array where phase is bound to [-pi, pi].
    """

    # calc and return
    shift = np.pi

    return (phase + shift) % (2 * np.pi) - shift


# noinspection SpellCheckingInspection,PyPep8Naming
def calculate_scale_correlation_function(
    sti_start: float,
    sti_end: float,
    dsti: float,
    filter_order: int,
    this,
    sti_norm_correction_fcxn,
):
    """Calculates the scale-correlation correlogram.

    `sti` is the ratio of temporal scales between a first and second filter.

    The S-SCF correlogram goes as

        sscf(sti) = 1/(2pi) int_reals H(w) H(w sti) dw

    where `H` is the Fourier transfer function. A normalization correction
    is required for inline-slope and -curvature filters because the
    leading `s` or `s^2` is attached before the integral is taken, which
    in turn produces an errant `sti` or `sti^2` scale factor.

    Parameters
    ----------
    sti_start: float
        Starting scale ratio, typically 1.0
    sti_end: float
        Ending scale ratio
    dsti: float
        Step size for scale-ratio increments
    filter_order: int
        The order of the filter
    this
        Handle to the calling filter-signature object
    sti_norm_correction_fcxn
        Function of `sti`: f(sti) -> float.

    Returns
    -------
    np.ndarray
        Array [n, 2] with cols (stigma_axis, sscf-norm-value).
    """

    # defs
    tau_base = 1.0
    stigma_axis = np.arange(sti_start, sti_end + dsti, dsti)

    # setup sscf correlogram array [n, 2]
    sscf_norm = np.zeros((stigma_axis.shape[0], 2))
    sscf_norm[:, 0] = stigma_axis

    # fetch callable H function and base-line scale (kh0)
    H = this.make_callable_transfer_function(filter_order)
    kh0 = this.design.autocorrelation_peak_and_stride_values(
        tau_base, filter_order
    )["kh0"]

    # iter on stigma, evaluating the integral over the xfer fcxn each step
    for i, sti in enumerate(stigma_axis):

        # integral eval
        sscf_norm[i, 1] = (
            (np.sqrt(sti) / kh0)
            * 2.0
            * np.real(
                integrate.quad(lambda f: H(f) * H(-f * sti), 0.0, np.inf)
            )[0]
            * sti_norm_correction_fcxn(sti)
        )

    return sscf_norm


# noinspection SpellCheckingInspection,PyPep8Naming
class SpectraColumns(Enum):
    """
    Names for the columns in the panel returned by *.generate_spectra(..)
    in the figure signatures.
    """

    FREQ = 0
    GAIN = 1
    PHASE = 2
    GROUPDELAY = 3



"""
-------------------------------------------------------------------------------

Analog multistage-box level filter design:

Captures temporal, spectral and autocorrelation traits of this filter.

Wavenumbers are calculated numerically in Python in `calculate_wavenumber.py`.

-------------------------------------------------------------------------------
"""

import numpy as np
import sys

from irides.resources.containers import points
from irides.tools import design_tools


# noinspection SpellCheckingInspection
def design_id() -> str:
    """
    Since `design` modules are passed around as objects in this framework,
    a `design_id` gives a unique name.

    Returns
    -------
    str
        The id of this design.
    """

    return "mbox-level"


def design_type() -> design_tools.FilterDesignType:
    """This is a level filter"""

    return design_tools.FilterDesignType.LEVEL


# noinspection SpellCheckingInspection
def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns valid filter orders for this filter design.

    Parameters
    ----------
    strict: bool
        When true, only orders where h(0)=0 are included.

    Returns
    -------
    np.ndarray
        Array of valid filter orders
    """

    return design_tools.get_valid_filter_orders(
        design_tools.FilterDesignType.LEVEL, strict
    )


"""
-------------------------------------------------------------------------------

Filter-specific definitions

-------------------------------------------------------------------------------
"""


# noinspection SpellCheckingInspection,PyPep8Naming
def convert_tau_to_T(tau: float) -> float:
    """T = 2 tau.

    `tau` is a temporal scale, `T` is a parameter of the m-box filter.
    The first moment of an m-box level filter is M1 = tau = T / 2.
    Thus: T = 2 x tau.
    """

    return 2.0 * tau


# noinspection SpellCheckingInspection,PyPep8Naming
def convert_T_to_tau(T: float) -> float:
    """tau = T / 2.

    For an m-box level filter, M1 = tau = T / 2.
    Thus: tau = T / 2.
    """

    return T / 2.0


"""
-------------------------------------------------------------------------------

Standard features

-------------------------------------------------------------------------------
"""


# noinspection PyUnusedLocal
def unit_peak_time(filter_order: int, dt=0.0) -> float:
    """
    Unit in this case refers to the first moment = 1, so T = 2.
    Thus, the unit-peak time is 1.

    Parameters
    ----------
    filter_order: int
        Order of the filter.
    dt: float
        Grid step of continuous-time discretization (default=0.0).

    Returns
    -------
    float
        The unit peak time of a multistage box level filter.
    """

    # invariant of the design
    t_peak = 1.0 - dt

    return t_peak


# noinspection SpellCheckingInspection
def unit_peak_value(filter_order: int, dt=0.0) -> float:
    """
    The value of impulse response h(t_peak; T = 2). While perhaps there is an
    analytic expression for this peak value, here the results from Mathematica
    are tabulated.

    Only order 1-8 are available.

    Parameters
    ----------
    filter_order: int
        Order of the filter.
    dt: float
        Grid step of continuous-time discretization (default=0.0).

    Returns
    -------
    float
        The unit peak value of the impulse response.
    """

    # unit peak values (from Mathematica mbox_level.nb,
    # with discrete corrections from python)
    unit_peak_values = {
        1: 1.0 / 2.0,
        2: 1.0 * (1.0 - dt),
        3: 9.0 / 8.0,
        4: 4.0 / 3.0,
        5: 575.0 / 384.0,
        6: 33.0 / 20.0,
        7: 41209.0 / 23040.0,
        8: 604.0 / 315.0,
    }

    # check and return
    valid_filter_orders = get_valid_filter_orders()
    if filter_order in valid_filter_orders:
        return unit_peak_values[filter_order]
    else:
        return 0.0


# noinspection SpellCheckingInspection,PyPep8Naming
def peak_value_coordinate(tau: float, filter_order: int, dt=0.0):
    """
    Bundles the peak time and value, scaled by T, into a point container.

    Parameters
    ----------
    tau: float
        First moment of the filter.
    filter_order: int
        Order of the filter.
    dt: float
        Grid step of continuous-time discretization (default=0.0).

    Returns
    -------
    ContinuousTimePoint
        Object with .time and .value fields.
    """

    # construct
    T = convert_tau_to_T(tau)
    T_cal = 2.0
    peak_coord = points.ContinuousTimePoint(
        unit_peak_time(filter_order, dt) * (T / T_cal),
        unit_peak_value(filter_order, dt) / (T / T_cal),
    )

    # return
    return peak_coord


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def moment_value_generator(moment: int, tau: float = 1.0, dt: float = 0.0):
    """
    The unit moments here are for T/2 = 1, so that M1 = 1.

    Returns an anonymous function with the single argument (order: int),
    where order is a `filter_order`. The returned function captures the `tau`
    and `dt` arguments in this calling function.

    Note that the generated functions are not necy purely continuous time.
    Instead, they bridge continuous and discrete time by applying order-dt
    corrections to the moments. This is required because the CT series are
    in fact implemented in DT. Use dt=0.0 to recover pure CT moments.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # generator defs with dt capture
    # noinspection PyUnusedLocal
    def m0_gen(order: int):
        return 1.0

    # noinspection PyUnusedLocal
    def m1_gen(order: int):
        return tau - dt / 2.0

    def m2_gen(order: int):
        return ((3.0 * order + 1.0) / (3.0 * order)) * np.power(
            tau, 2
        ) - dt / 2.0

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection,PyPep8Naming
def autocorrelation_peak_and_stride_values(
    tau: float, filter_order: int
) -> dict:
    """
    The S-ACF parameters are all numerically determined. Here, the sacf(lag=0)
    and sacf(lag=xi_5pct) = 0.05 values are available for T = 2.

    The peak value of the S-ACF, sacf(lag=0), scales as 1 / T. The
    5% stride value xi_5pct, scales as T.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Peak value (scaled by 1/T) and 5% pct stride (scaled by T)
    """

    # convert 1st moment to T
    T_input = convert_tau_to_T(tau)

    # values for T = 2
    T_cal = 2.0
    unit_sacf_values = {
        1: {"kh0": 1.0 / 2.0, "xi_5pct": 1.90},
        2: {"kh0": 2.0 / 3.0, "xi_5pct": 1.4152},
        3: {"kh0": 33.0 / 40.0, "xi_5pct": 1.15309},
        4: {"kh0": 302.0 / 315.0, "xi_5pct": 0.998982},
        5: {"kh0": 78095.0 / 72576.0, "xi_5pct": 0.893614},
        6: {"kh0": 655177.0 / 554400.0, "xi_5pct": 0.815806},
        7: {"kh0": 189597667.0 / 148262400.0, "xi_5pct": 0.755319},
        8: {"kh0": 2330931341.0 / 1702701000.0, "xi_5pct": 0.706554},
    }
    zero_values = {"kh0": 0.0, "xi_5pct": 0.0}

    # check and return
    if filter_order in unit_sacf_values.keys():
        ans = unit_sacf_values[filter_order]
        ans["kh0"] *= T_cal / T_input
        ans["xi_5pct"] *= T_input / T_cal
    else:
        ans = zero_values

    return ans


# noinspection SpellCheckingInspection,PyPep8Naming
def scale_correlation_stride_and_residual_values(
    tau: float, filter_order: int
) -> dict:
    """Scale decorrelation strides and residual correlations.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Scale decorrelation (scales as `tau`) and residual correlation
        (independent of `tau`).
    """

    # values from Mathematica
    unit_sscf_values = {
        1: {"sti_5pct": 400.000, "residual": 0.05},
        2: {"sti_5pct": 9.65489, "residual": 0.05},
        3: {"sti_5pct": 4.60289, "residual": 0.05},
        4: {"sti_5pct": 3.36003, "residual": 0.05},
        5: {"sti_5pct": 2.81947, "residual": 0.05},
        6: {"sti_5pct": 2.51131, "residual": 0.05},
        7: {"sti_5pct": 2.30873, "residual": 0.05},
        8: {"sti_5pct": 2.16416, "residual": 0.05},
    }

    # fetch values for order and scale
    sscf_values = unit_sscf_values[filter_order]
    sscf_values["sti_5pct"] *= tau

    # return
    return sscf_values


# noinspection SpellCheckingInspection,PyPep8Naming
def full_width_generator(tau: float = 1.0, dt: float = 0.0):
    """
    Returns anonymous function for the full-width, given that T/2 = 1.
    `tau` and `dt` are captured here, and the anonymous function has one
    argument, (order: int), the filter order.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    def fw_gen(order: int):
        return 2.0 * np.sqrt(1.0 / (3.0 * order) - np.power(dt / tau, 2)) * tau

    return fw_gen


# noinspection PyUnusedLocal
def gain_at_dc(tau: float) -> float:
    """Filter gain at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return 1.0


# noinspection PyUnusedLocal
def phase_at_dc(tau: float) -> float:
    """Filter phase at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return 0.0


def group_delay_at_dc(tau: float) -> float:
    """Filter group delay at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return tau


# noinspection SpellCheckingInspection,PyPep8Naming
def cutoff_frequency(tau: float, filter_order: int) -> float:
    """
    At w_cutoff,

        |Hmbox(w_cutoff; T, m)|^2 = 1/2.

    The solution is not analytic so the values for order 1-8 are reported here
    for T_cal = 2.

    The scaling is clear when written as this:

        g^2(T_cal wc_cal) = g^2(T wc') = 1/2,

    so,

        wc' = wc_cal Tcal / T.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter  (note: user is required to validate that
        the order is supported).

    Returns
    -------
    float
        Frequency at cutoff (radians).
    """

    # convert
    T = convert_tau_to_T(tau)

    # the calibration temporal scale
    T_cal = 2.0

    # cutoff freqs (from Mathematica mbox_level.nb, T=2)
    cutoff_frequencies = {
        1: 1.39,
        2: 2.00,
        3: 2.47,
        4: 2.86,
        5: 3.20,
        6: 3.51,
        7: 3.80,
        8: 4.06,
    }

    # check and return
    return cutoff_frequencies[filter_order] * T_cal / T


# noinspection PyUnusedLocal
def gain_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The cutoff gain is always sqrt(1/2). The purpose of this function is
    to adhere to the phase- and group-delay-cutoff api so that any of the
    gain/phase/group-delay values @ cutoff can be determined uniformly.
    """

    return np.sqrt(1.0 / 2.0)


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def phase_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The phase at cutoff frequency `wc`. Phase at cutoff is independent
    of temporal scale `tau`, and this being a multistage box filter,
    the phase is also independent of filter order.

    Parameters
    ----------
    tau: float
        Unused but present for a consistent api
    filter_order: int
        The order of the filter (unused).

    Returns
    -------
    float
        The phase at the cutoff frequency.
    """

    T_input = convert_tau_to_T(tau)
    T_cal = 2.0
    tau_cal = convert_T_to_tau(T_cal)
    w_cutoff = cutoff_frequency(tau_cal, filter_order)

    return -w_cutoff * T_input / T_cal


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def group_delay_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The group delay for a multistage box filter is T/2 independent of the
    filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    float
    """

    # convert
    T = convert_tau_to_T(tau)

    return T / 2.0


# noinspection SpellCheckingInspection
def wireframe(tau: float, filter_order: int):
    """Wireframe timepoint for this level filter.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter (unused)

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    return design_tools.generate_level_wireframe_from_spectra(
        sys.modules[__name__], tau, filter_order
    )


# noinspection SpellCheckingInspection
def wavenumber(filter_order: int) -> float:
    """Report the wavenumber for this filter as a function of order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The wavenumber.
    """

    wavenumbers = {
        1: 0.76907485,
        2: 0.79476859,
        3: 0.76418367,
        4: 0.72115753,
        5: 0.68203083,
        6: 0.65227649,
        7: 0.63221627,
        8: 0.62143716,
    }

    return wavenumbers[filter_order]


# noinspection SpellCheckingInspection
def uncertainty_product(filter_order: int) -> dict:
    """Dict of TW, BW, and UCP for this filter order.

    Parameters
    ----------
    filter_order: int
        The order of the filter

    Returns
    -------
    dict
        TW is temporal width, BW is spectral width, and UCP is the
        uncertainty product.
    """

    uncertainty_products = {
        1: {"TW": 0.57735, "BW": 427.856, "UCP": 247.023},
        2: {"TW": 0.316228, "BW": 1.73205, "UCP": 0.547723},
        3: {"TW": 0.249096, "BW": 2.0226, "UCP": 0.503822},
        4: {"TW": 0.212513, "BW": 2.35858, "UCP": 0.50123},
        5: {"TW": 0.18846, "BW": 2.65682, "UCP": 0.500706},
        6: {"TW": 0.171088, "BW": 2.92524, "UCP": 0.500474},
        7: {"TW": 0.157781, "BW": 3.17111, "UCP": 0.500342},
        8: {"TW": 0.147166, "BW": 3.39929, "UCP": 0.500258},
    }

    return uncertainty_products[filter_order]


"""
-------------------------------------------------------------------------------

Filter-specific features

-------------------------------------------------------------------------------
"""


# noinspection SpellCheckingInspection,PyPep8Naming
def first_lobe_frequency_and_gain_values(
    tau: float, filter_order: int
) -> tuple:
    """
    Returns the radial frequency w for the gain peak of the first
    gain lobe, along with the gain value.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    tuple
        (w_1st_lobe, gain_1st_lobe)
        w_1st_lobe: Radial frequency for gain peak of first gain lobe
        gain_1st_lobe: Gain at w_1st_lobe
    """

    T = convert_tau_to_T(tau)
    w_1st_lobe = 3.0 * np.pi / (T / filter_order)
    gain_1st_lobe = np.power(2.0 / (3.0 * np.pi), filter_order)

    return w_1st_lobe, gain_1st_lobe



"""
-------------------------------------------------------------------------------

Simple containers to structure points.

-------------------------------------------------------------------------------
"""

import numpy as np


class ContinuousTimePoint(object):
    """
    Container for a continuous-time time point and associated
    series value.
    """

    def __init__(self, time=None, value=None):
        """Constructor takes time and value; defaults to None."""

        self.time = time
        self.value = value


class ContinuousTimePoints(object):
    def __init__(self, time: np.ndarray, value: np.ndarray):

        self.points = np.array([time, value]).T

    def times(self) -> np.ndarray:

        return self.points[:, 0]

    def values(self) -> np.ndarray:

        return self.points[:, 1]


class ContinuousFrequencyPoint(object):
    """
    Container for a continuous-frequency frequency point
    and associated value.
    """

    def __init__(self, frequency=None, value=None):
        """Constructor takes time and value; defaults to None."""

        self.frequency = frequency
        self.value = value


class DiscreteTimePoint(object):
    """
    Container for a discrete-time time point and associated
    series value.
    """

    def __init__(self, index=None, value=None):
        """Constructor takes index and value; defaults to None."""

        self.index = index
        self.value = value



"""
-------------------------------------------------------------------------------

Analog bessel inline curvature filter design:

Captures the design details for this filter. Design details are available
in Mathematica notebook bessel_inline_curvature.nb.

-------------------------------------------------------------------------------
"""

import numpy as np
from enum import Enum

from irides.resources.core_enumerations import FilterDesignType
from irides.resources.containers.wireframes import WireframeContinuousTime


class StageTypes(Enum):
    """Defines the set of possible stage types"""

    EMA = 0
    DOSC = 1


def get_valid_filter_orders_for_level_filters() -> np.ndarray:
    """Returns array of valid filter orders across all level filters"""

    return np.array([1, 2, 3, 4, 5, 6, 7, 8], dtype=int)


# noinspection SpellCheckingInspection,PyUnusedLocal
def get_minimum_valid_filter_order(
    filter_type: FilterDesignType, strict: bool = False
) -> int:
    """Returns minimum valid filter order.

    Strict orders have the trait that h(0) = 0, regardless of filter design
    type.

    Parameters
    ----------
    filter_type: FilterDesignType
        Indicates level, slope, or curve filter type.
    strict: bool
        Select strict or not-strict (loose) lower bound

    Returns
    -------
    int
        Minimum valid filter order.
    """

    soln_dict = {
        FilterDesignType.LEVEL.value: 1,
        FilterDesignType.SLOPE.value: 2,
        FilterDesignType.CURVE.value: 3,
    }

    minimum_valid_order = soln_dict[filter_type.value]

    if strict:
        minimum_valid_order += 1

    return minimum_valid_order


# noinspection SpellCheckingInspection
def get_valid_filter_orders(
    filter_type: FilterDesignType, strict: bool = False
) -> np.ndarray:
    """Returns array of valid filter orders based on f-type and strictness

    Parameters
    ----------
    filter_type: FilterDesignType
        Indicates level, slope, or curve filter type.
    strict: bool
        If true, return only higher orders, removing lower orders that
        do not qualify as suitable filters.

    Returns
    --------
    np.ndarray:
        Array of valid filter orders
    """

    filter_orders = get_valid_filter_orders_for_level_filters()
    minimum_valid_order = get_minimum_valid_filter_order(filter_type, strict)

    filter_orders = np.delete(
        filter_orders, np.where(filter_orders < minimum_valid_order)
    )

    return filter_orders


# noinspection SpellCheckingInspection,PyUnusedLocal
def wireframe_moment_value_generator(
    design, moment: int, tau: float, dt: float
):
    """Returns M0, M1, or M2 moment generators given a filter-operation type."""
    moment_generating_functions = {
        FilterDesignType.LEVEL: {
            0: lambda order: 1.0,
            1: lambda order: tau,
            2: lambda order: np.power(tau, 2),
        },
        FilterDesignType.SLOPE: {
            0: lambda order: 0.0,
            1: lambda order: -1.0,
            2: lambda order: -(
                design.wireframe(tau, order).timepoints[0]
                + design.wireframe(tau, order).timepoints[1]
            ),
        },
        FilterDesignType.CURVE: {
            0: lambda order: 0.0,
            1: lambda order: 0.0,
            2: lambda order: 2.0,
        },
    }

    return moment_generating_functions[design.design_type()][moment]


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def generate_level_wireframe_from_spectra(
    design, tau: float, filter_order: int
) -> WireframeContinuousTime:
    """Calculates and constructs a level-type wireframe object."""

    return WireframeContinuousTime(FilterDesignType.LEVEL, np.array([tau]))


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_slope_wireframe_from_spectra(
    design, tau: float, filter_order: int
) -> WireframeContinuousTime:
    """Calculates and constructs a slope-type wireframe object.

    Parameters
    ----------
    design:
        Reference to a design module.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    # frequency coordinates
    w_peak = design.frequency_of_peak_gain(tau, filter_order)
    phase_peak = design.phase_at_peak_frequency(tau, filter_order)

    # convert frequency of peak gain to a (co)sinusoidal period
    T_peak = 2.0 * np.pi / w_peak

    # the slope wireframe is one-half of T_peak
    T_wf = T_peak / 2.0

    # the positive timepoint is calculated thru the argument of the cosine
    #   cos(w_peak t_pos + phase_peak) --> cos(0) = 1
    t_pos = -phase_peak / w_peak

    # offset by T_wf to find t_neg
    t_neg = t_pos + T_wf

    # construct and return a wireframe container
    return WireframeContinuousTime(
        FilterDesignType.SLOPE, np.array([t_pos, t_neg])
    )


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_curvature_wireframe_from_spectra(
    design, tau: float, filter_order: int
) -> WireframeContinuousTime:
    """Calculates and constructs a curvature-type wireframe object.

    Parameters
    ----------
    design:
        Reference to a design module.
    filter_order: int
        The order of the filter.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    # frequency coordinates
    w_peak = design.frequency_of_peak_gain(tau, filter_order)
    phase_peak = design.phase_at_peak_frequency(tau, filter_order)

    # convert frequency of peak gain to a (co)sinusoidal period
    T_peak = 2.0 * np.pi / w_peak

    # the curvature wireframe is the same as T_peak
    T_wf = T_peak

    # the negative (center) timepoint is calculated thru
    # the argument of the cosine
    t_pos_l = -phase_peak / w_peak

    # the left and right positive timepoints are symmetrically offset from
    # the center timepoint
    t_neg = t_pos_l + T_wf / 2.0
    t_pos_r = t_pos_l + T_wf

    # construct and return a wireframe container
    return WireframeContinuousTime(
        FilterDesignType.CURVE, np.array([t_pos_l, t_neg, t_pos_r])
    )



"""
-------------------------------------------------------------------------------

Digital damped oscillator:

Implements the temporal signature of a damped-oscillator filter.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.digital import damped_oscillator as dsgn_dig_dosc
from irides.resources.containers.discrete_sequence import DiscreteSequence


# noinspection SpellCheckingInspection
def generate_impulse_response(
    n_start: int, n_end: int, zplane_poles: np.ndarray
) -> DiscreteSequence:
    """Creates a discrete-time impulse response of the damped oscillator.

    In the case that Im(zplane_poles) == 0, the 2nd-order pole-ema solution
    is returned.

    Parameters
    ----------
    n_start: int
        Start sample.
    n_end: int
        End sample (exclusive).
    zplane_poles: np.ndarray
        Coordinates of complex-conjugate pole pair.

    Returns
    -------
    DiscreteSequence
        Contains n-axis and h-axis
    """

    # setup a sequence template
    ds = DiscreteSequence(n_start, n_end)

    # pull n-axis, n == 0 index, and alias v-axis
    i_n_zero = ds.i_n_zero
    n_0plus = ds.n_axis[i_n_zero:]
    hn = ds.v_axis

    # convert pole to polar form
    i_select = (
        0
        if zplane_poles[0] == zplane_poles[1]
        else np.argwhere(np.imag(zplane_poles) > 0)[0][0]
    )
    gain_adj_v, r_v, phi_v = dsgn_dig_dosc.gain_adjustment(
        zplane_poles, return_polar_coords=True
    )
    gain_adj = gain_adj_v[i_select]
    r = r_v[i_select]
    phi = phi_v[i_select]

    # compute h[n]
    if np.abs(np.imag(zplane_poles[0])) > 0.0:
        hn[i_n_zero:] = (
            gain_adj
            / np.sin(phi)
            * np.sin((n_0plus + 1) * phi)
            * np.power(r, n_0plus)
        )
    else:
        hn[i_n_zero:] = (
            np.cos(phi * n_0plus)
            * np.power(1.0 - r, 2)
            * (n_0plus + 1)
            * np.power(r, n_0plus)
        )

    # return
    return ds



"""
-------------------------------------------------------------------------------

A container for cartesian_poles. Ctor consumes Cartesian form and the
interface provides for Cartesian or polar return values.

-------------------------------------------------------------------------------
"""

import numpy as np


class PoleContainer(object):
    """Container for nd-array of complex-valued cartesian_poles."""

    def __init__(self, cartesian_poles: np.ndarray):
        """Ctor to the PoleContainer.

        Parameters
        ----------
        cartesian_poles: np.ndarray
            An nd-array of complex-valued poles.
            dtype=complex (accepts float)
        """

        if isinstance(cartesian_poles, np.ndarray):
            self._cartesian_poles = cartesian_poles
        else:
            msg = "ctor argument must be an ndarray of pole coords."
            raise RuntimeError(msg)

    def cartesian_poles(self) -> np.ndarray:
        """Retrieve cartesian_poles in Cartesian form.

        Returns
        -------
        np.ndarray
            Poles in Cartesian form
        """

        return self._cartesian_poles

    def polar_poles(self, from_positive_real_axis: bool = True) -> np.ndarray:
        """Retrieve cartesian_poles in polar form,

        Parameters
        ----------
        from_positive_real_axis: bool
            Whether or not to reference polar angle to pos (T) or neg (F) axis

        Returns
        -------
        np.ndarray
            Poles in polar form
        """

        # manage real-axis angle
        sgn = 1.0 if from_positive_real_axis else -1.0

        # convert Cartesian to polar
        radius = np.abs(self._cartesian_poles)
        angle = np.arctan2(
            np.imag(self._cartesian_poles), sgn * np.real(self._cartesian_poles)
        )

        # pack and return
        return np.array([radius, angle]).T



"""
-------------------------------------------------------------------------------

Container for wireframes.

-------------------------------------------------------------------------------
"""

import numpy as np
from dataclasses import dataclass

from irides.resources.core_enumerations import FilterDesignType


# noinspection SpellCheckingInspection
@dataclass
class WireframeContinuousTime(object):
    """Container for level, slope, or curvature wireframes.

    type: FilterDesignType -- {LEVEL|SLOPE|CURVE}
    timepoints: np.ndarray -- 1, 2, or 3 timepoints
    weights: np.ndarray -- timepoint weights so that the * gain is unity
    change_interval: float -- timespan over which the wireframe covers
    mid_point: float -- mean of the timepoints
    """

    # locals
    type = FilterDesignType.UNKNOWN
    timepoints = np.array([])
    weights = np.array([])
    change_interval = None
    mid_point = None

    def __init__(
        self, design_type: FilterDesignType, wireframe_timepoints: np.ndarray
    ):
        """Initializes an instance.

        Parameters
        ----------
        design_type: FilterDesignType
            Type of wireframe that this instance refers to
        wireframe_timepoints: np.ndarray
            Wireframe timepoints
        """

        # validate type
        if design_type == FilterDesignType.UNKNOWN:
            raise RuntimeError("Wireframe cannot be `unknown`.")

        # capture type
        self.type = design_type

        # capture timepoints
        self.timepoints = np.asarray(wireframe_timepoints)

        # map to calculation functions
        set_weights = set_weights_functions[design_type]
        change_function = change_functions[design_type]
        midpoint_function = midpoint_functions[design_type]

        # exec
        set_weights(self)
        self.change_interval = change_function(self)
        self.mid_point = midpoint_function(self)

        # validate config
        if design_type == FilterDesignType.LEVEL:
            validate_check_or_die(self, 1, "level")
        if design_type == FilterDesignType.SLOPE:
            validate_check_or_die(self, 2, "slope")
        if design_type == FilterDesignType.CURVE:
            validate_check_or_die(self, 3, "curvature")


"""
-------------------------------------------------------------------------------

Supporting functions, specific to wireframe type

-------------------------------------------------------------------------------
"""


def set_level_weights(self):
    self.weights = np.zeros(1)
    self.weights[0] = 1.0


def set_slope_weights(self):
    h = self.timepoints[1] - self.timepoints[0]
    self.weights = np.zeros(2)
    self.weights[0] = 1.0 / h
    self.weights[1] = -1.0 / h


def set_curvature_weights(self):
    h = self.timepoints[2] - self.timepoints[0]
    self.weights = np.zeros(3)
    self.weights[0] = 1.0 / np.power(h / 2.0, 2)
    self.weights[1] = -2.0 / np.power(h / 2.0, 2)
    self.weights[2] = 1.0 / np.power(h / 2.0, 2)


set_weights_functions = {
    FilterDesignType.LEVEL: set_level_weights,
    FilterDesignType.SLOPE: set_slope_weights,
    FilterDesignType.CURVE: set_curvature_weights,
}


change_functions = {
    FilterDesignType.LEVEL: lambda self: 0.0,
    FilterDesignType.SLOPE: lambda self: self.timepoints[1]
    - self.timepoints[0],
    FilterDesignType.CURVE: lambda self: self.timepoints[2]
    - self.timepoints[0],
}

midpoint_functions = {
    FilterDesignType.LEVEL: lambda self: self.timepoints[0],
    FilterDesignType.SLOPE: lambda self: np.mean(
        [self.timepoints[0], self.timepoints[1]]
    ),
    FilterDesignType.CURVE: lambda self: np.mean(
        [self.timepoints[0], self.timepoints[2]]
    ),
}


def validate_check_or_die(self, count: int, label: str):
    if self.timepoints.shape[0] != count:
        raise RuntimeError(
            "{0} filter must have {1} timepoints".format(label, count)
        )



"""
-------------------------------------------------------------------------------

Container for a discrete-time sequence wherein the n-axis is int-type while
value(s) are float-type.

-------------------------------------------------------------------------------
"""

from __future__ import annotations
import numpy as np
import copy


class DiscreteSequence(object):
    """Container of a discrete-time sequence"""

    def __init__(
        self, n_start: int, n_end: int, n_cols: int = 1, v_type: str = "float"
    ):
        """Initializes a discrete-time sequence.

        Parameters
        ----------
        n_start: int
            Start sample.
        n_end: int
            End sample (exclusive)
        n_cols: int
            Number of value columns (default = 1)
        """

        # initialize and save locally
        self._n_axis = np.arange(n_start, n_end, dtype="int64")
        self._v_axes = np.zeros((self._n_axis.shape[0], n_cols), dtype=v_type)
        self._step_size = 1

        # find index into n == 0 (if it exists)
        candidate = np.where(self._n_axis == 0)[0]
        self._i_n_zero = candidate[0] if candidate.shape[0] > 0 else None

    @classmethod
    def copy(cls, obj) -> DiscreteSequence:
        """Returns of deep copy of an instance of this object"""

        return copy.deepcopy(obj)

    @property
    def step_size(self) -> int:
        """Returns the step size of the sequence"""

        return self._step_size

    @property
    def len(self) -> int:
        """Returns the length of the sequence"""

        return self._n_axis.shape[0]

    @property
    def n_axis(self):
        """Returns the index axis."""

        return self._n_axis

    @property
    def i_n_zero(self):
        """Returns the index into n_axis where n == 0, or None"""

        return self._i_n_zero

    @property
    def v_axis(self):
        """Returns the zeroth column of values"""

        return self._v_axes[:, 0]

    @v_axis.setter
    def v_axis(self, new_v_axis_values):
        """Sets the zeroth column of values"""

        self._v_axes[:, 0] = new_v_axis_values

    @property
    def v_axes(self):
        """Returns the value axes"""

        return self._v_axes

    @v_axes.setter
    def v_axes(self, new_v_axes_values):
        """Sets the value axes (subject to throw if sizes don't match)"""

        self._v_axes = new_v_axes_values

    def to_ndarray(self):
        """Returns a single np.ndarray where n_axis (now float) is 1st col."""

        return np.hstack((self._n_axis[:, np.newaxis], self._v_axes))

    def index_at(self, n_value: int) -> int:
        """Returns index i such that ds.n_axis[i] == n_value"""

        candidate = np.where(self._n_axis == n_value)[0]
        return candidate[0] if candidate.shape[0] > 0 else None



"""
-------------------------------------------------------------------------------

Continuous-time test signals.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.resources.core_enumerations import TestSignalType
from irides.tools import impulse_response_builder_tools


# noinspection SpellCheckingInspection
def generate_test_signal(
    t_start: float,
    t_end: float,
    dt: float,
    signal_type: TestSignalType,
    alternating_frequency: float = 0.0,
) -> tuple:
    """Approximation of continuous-time test signal on a discretized time grid.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    signal_type: TestSignalType
        Type of test signal to construct.
    alternating_frequency: float
        Radial frequency of alternating signal, w: cos(w t) [optional]

    Returns
    -------
    tuple
        (np.ndarray, int)
        np.ndarray: A time series array (n, 2) having columns [t_axis, h_test].
        int: index to t=0 value (if it exists)
    """

    # create a time-series template
    # fmt: off
    ts, i_t0_plus = impulse_response_builder_tools.make_impulse_response_template(
        t_start, t_end, dt, n_cols=2
    )
    # fmt: on
    t_axis = ts[:, 0]

    # quick return if type is unkonwn
    if signal_type == TestSignalType.UNKNOWN:
        return ts, i_t0_plus

    # build lambda dict
    time_functions = {
        TestSignalType.IMPULSE: lambda t: 0.0,  # note: this requires a fixup
        TestSignalType.STEP: lambda t: 1.0,
        TestSignalType.RAMP: lambda t: t,
        TestSignalType.PARABOLA: lambda t: 0.5 * np.power(t, 2),
        TestSignalType.ALTERNATING: lambda t: np.cos(alternating_frequency * t),
        TestSignalType.WHITE_NOISE: lambda t: np.random.normal(0, 1, 1)[0],
    }

    # apply
    ts[:, 1] = np.array(list(map(time_functions[signal_type], t_axis)))

    # impulse fixup
    if signal_type == TestSignalType.IMPULSE:
        ts[i_t0_plus, 1] = 1.0 / dt

    # force causality
    ts[:i_t0_plus, 1] = 0.0

    # return
    return ts, i_t0_plus



"""
-------------------------------------------------------------------------------

Analog multistage-box inline curvature filter:

Implements the temporal, spectral, autocorrelation and pole/zero features
for this continuous-time inline-slope filter.

-------------------------------------------------------------------------------
"""

from typing import Callable
import numpy as np
import sys

from irides.design.analog import mbox_inline_curvature as design

from irides.filter_signatures.analog import mbox_level as f_sig_ref

from irides.tools import impulse_response_builder_tools
from irides.tools import transfer_function_builder_tools
from irides.tools import impulse_response_tools
from irides.tools import transfer_function_tools


# noinspection SpellCheckingInspection
def generate_impulse_response(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Creates a continuous-time-like impulse response for this filter.
    The implementation starts with a temporally scaled level impulse
    response and then takes the causal second difference.

    Only orders [3..8] are implemented.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        A time series array (n, 2) having columns [t_axis, h].
    """

    # start with a level-filter impulse response
    time_series = f_sig_ref.generate_impulse_response(
        t_start, t_end, dt, tau, filter_order
    )

    # fmt: off
    i_t_0plus = impulse_response_builder_tools. \
        compute_noncausal_symmetric_second_difference(
            time_series
        )
    # fmt: on

    # h(0) correction
    ht0_value = design.impulse_response_t0_value(tau, filter_order, dt)
    if i_t_0plus is not None:
        time_series[i_t_0plus, 1] = ht0_value

    return time_series


# noinspection SpellCheckingInspection
def generate_sacf_correlogram(
    xi_start: float,
    xi_end: float,
    dxi: float,
    tau: float,
    filter_order: int,
    normalize=False,
) -> np.ndarray:
    """
    Computes the system autocorrelation response from the correlation
    of the impulse response with itself.

    Parameters
    ----------
    xi_start: float
        Start lag, |xi_start| <= |xi_end|
    xi_end: float
        End lag
    dxi: float
        Lag increment
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.
    normalize: bool
        T/F whether to normalize the sacf spectrum so sacf(0) = 1.

    Returns
    -------
    np.ndarray
        System autocorrelation (n, 2) np.array w/ cols [xi, sacf-value].
    """

    # fetch impulse response
    ir = generate_impulse_response(0.0, xi_end, dxi, tau, filter_order)

    # calc sacf
    sacf = impulse_response_tools.calculate_auto_correlation_function(ir)

    # fetch kh0
    kh0 = design.autocorrelation_peak_and_stride_values(tau, filter_order, dxi)[
        "kh0"
    ]

    # post process
    # fmt: off
    sacf = transfer_function_builder_tools. \
        truncate_and_normalize_sacf_correlogram(
            sacf, kh0, normalize, xi_start
        )
    # fmt: on

    return sacf


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_transfer_function(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Computes the transfer function H_curv(f) based on the
    underlying reference filter H_ref(f):

        H_curv(tau s) = s^2 H_ref(tau s).

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        (n, 2), dtype=complex
        The transfer function with cols [f_axis, H_curv(f)].
    """

    # compute reference transfer function
    H = f_sig_ref.generate_transfer_function(
        f_start, f_end, df, tau, filter_order
    )

    # convert to inline curvature
    # fmt: off
    transfer_function_builder_tools. \
        apply_leading_jw_coefficient_to_transfer_function(H, 2)
    # fmt: on

    # return
    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def make_callable_transfer_function(
    filter_order: int,
) -> Callable[[float, float], complex]:
    """Returns a callable function with signature (f, tau) -> complex.

    Constructs H(jw) = (jw)^2 H_ref(tau jw).

    Parameters
    ----------
    filter_order: int
        The filter order, which is captured by the definition herein.

    Returns
    -------
    Callable[[float, float], complex]
        A callable function (f: float, tau: float) -> complex
    """

    # fmt: off
    power = 2
    return transfer_function_builder_tools. \
        make_callable_transfer_function_with_leading_jw_power_term(
            f_sig_ref, filter_order, power
        )
    # fmt: on


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_sscf_correlogram(
    sti_start: float, sti_end: float, dsti: float, filter_order: int
) -> np.ndarray:
    """
    Computes the normalized system scale-correlation correlogram from the
    filter's transfer function.

    Note: inline-curvature filters have the transfer function

        s^2 H(tau s)

    and the sscf integral covers

        int_reals s^4 H(s) H(-s sti) ds.

    The way the code is structured, `s` is attached to `H(s)` before this
    integral is evaluated, creating

        int_reals s^2 (s sti)^2 H(s) H(-s sti) ds.

    Therefore, the integral needs to be corrected by 1 / sti^2.

    Parameters
    ----------
    sti_start: float
        Start scale, |sti_start| <= |sti_end|
    sti_end: float
        End scale
    dsti: float
        Scale increment
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        System scale-correlation normalized (n, 2)
        np.array w/ cols [sti, sscf-value].
    """

    # correction function
    sti_norm_correction_fcxn = lambda sti: 1.0 / np.power(sti, 2)

    # return sscf-norm np-array
    return transfer_function_tools.calculate_scale_correlation_function(
        sti_start,
        sti_end,
        dsti,
        filter_order,
        sys.modules[__name__],
        sti_norm_correction_fcxn,
    )


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_phase_spectrum(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Computes the phase of H_curv(f) based on the underlying reference
    filter H_ref(f):

        angle_H_curv(f) = pi + angle_H_ref(f).

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        (n, 2), dtype=complex
        The phase spectrum with cols [f_axis, angle_H_curv(f)].
    """

    # compute reference transfer function
    H = f_sig_ref.generate_phase_spectrum(f_start, f_end, df, tau, filter_order)

    # convert to inline slope
    # fmt: off
    transfer_function_builder_tools. \
        apply_leading_jw_coefficient_to_phase_spectrum(H, 2)
    # fmt: on

    # return
    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_spectra(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    """Generate gain, phase and group-delay spectra.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        Spectra panel [n_pts, 4] with cols [f, g(f), phase(f), grpdly(f)].
    """

    H = generate_transfer_function(f_start, f_end, df, tau, filter_order)

    phase_H = generate_phase_spectrum(f_start, f_end, df, tau, filter_order)

    # return spectra
    return transfer_function_tools.calculate_spectra(H, phase_H)


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_wireframe(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    """Returns a wireframe equivalent for this filter.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    WireframeContinuousTime
        A wireframe container.
    """

    # fetch the wireframe, create a timeseries template
    wf = design.wireframe(tau, filter_order)
    ts = impulse_response_builder_tools.discretize_wireframe_onto_time_grid(
        t_start, t_end, dt, wf
    )

    return ts



"""
-------------------------------------------------------------------------------

Analog damped oscillator:

Implements the temporal and spectral signatures of a continuous-time
damped-oscillator filter.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.analog import damped_oscillator as design
from irides.tools import impulse_response_builder_tools
from irides.tools import impulse_response_tools
from irides.tools import transfer_function_tools


# noinspection SpellCheckingInspection
def generate_impulse_response(
    t_start: float, t_end: float, dt: float, reference_pole: complex, tau: float
) -> np.ndarray:
    r"""Creates continuous-time-like impulse response of the damped oscillator.

    The angle of inclination of the reference pole with respect to the negative
    real axis governs which expression is used to construct the impulse
    response.

    In the general case, 0 < |angle(reference_pole)|, so

        h(t) = wn csc(theta) sin( sin(theta) wn t ) exp( -cos(theta) wn t ) u(t)

    but for angle(reference_pole) == 0,

        h(t) = wn^2 t exp( -wn t ) u(t).

    Here,

        wn = |reference_pole|
        theta = angle(reference_pole)

    where theta is respect to the negative real axis. Note that /reference_pole/
    is expected to be cartesian, and accordingly its angle will be wrt to
    positive real axis, so an adjustment is made.

    Scaling by tau is done through the law

        h(t) ->  tau^{-1} h(t/tau) .

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    reference_pole: complex
        Cartesian coordinates of pole.
    tau: float
        First moment of the filter.

    Returns
    -------
    np.ndarray
        A time series array (n, 2) having columns [t_axis, h_dosc].
    """

    # make time series, an (n, 2) matrix
    time_series = np.zeros((np.arange(t_start, t_end + dt, dt).shape[0], 2))

    # make time series template
    (
        time_series,
        i_t_0plus,
    ) = impulse_response_builder_tools.make_impulse_response_template(
        t_start, t_end, dt
    )

    # alias
    t_axis = time_series[:, 0]
    t_0plus = t_axis[i_t_0plus:]
    h = time_series[:, 1]

    # from /reference_pole/ extract wn and theta
    wns, thetas = design.extract_wn_and_theta(np.array([reference_pole]))
    wn = wns[0]
    theta = thetas[0]

    # branch on theta
    if theta == 0.0:

        h[i_t_0plus:] = (
            np.power(wn, 2) * (t_0plus / tau) * np.exp(-wn * t_0plus / tau)
        )

    else:

        h[i_t_0plus:] = (
            wn
            / (tau * np.sin(theta))
            * np.sin(np.sin(theta) * wn * t_0plus / tau)
            * np.exp(-np.cos(theta) * wn * t_0plus / tau)
        )

    # return
    return time_series


# noinspection SpellCheckingInspection
def generate_sacf_correlogram(
    xi_start: float,
    xi_end: float,
    dxi: float,
    reference_pole: complex,
    tau: float,
    normalize=False,
) -> np.ndarray:
    """Creates S-ACF response by impulse-response correlation with itself.

    Parameters
    ----------
    xi_start: float
        Start lag, |xi_start| <= |xi_end|
    xi_end: float
        End lag
    dxi: float
        Lag increment
    reference_pole: complex
        Cartesian coordinates of pole.
    tau: float
        First moment of the filter.
    normalize: bool
        T/F whether to normalize the sacf spectrum so sacf(0) = 1.

    Returns
    -------
    np.ndarray
        System autocorrelation (n, 2) np.array w/ cols [xi, sacf-value].
    """

    # fetch impulse response
    ir = generate_impulse_response(0.0, xi_end, dxi, reference_pole, tau)

    # calc sacf
    sacf = impulse_response_tools.calculate_auto_correlation_function(ir)

    # truncate at xi_start
    sacf = sacf[sacf[:, 0] >= xi_start - dxi / 2.0, :]

    # normalize by theoretic peak value if requested
    if normalize:
        kh0 = design.autocorrelation_peak_and_stride_values(
            np.array([reference_pole]), tau
        )["kh0"]
        sacf[:, 1] /= kh0

    return sacf


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_transfer_function(
    f_start: float, f_end: float, df: float, reference_pole: complex, tau: float
) -> np.ndarray:
    """
    Returns an np array, first column is the (cyclic) frequency axis,
    second column is the complex-valued transfer function.

    The defining transfer function is

                                 p p*
        H_dosc(tau s) = -------------------------
                        (tau s + p)(tau s + p*)

    In general, s = omega + j w, a complex value. Here, however, omega = 0,
    so /s/ is replaced by /j w/.

    Moreover, the api is for cyclic frequency /f/, measured in cycles per sec,
    but radial frequency /w/, measured in radians per sec, is used internally.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    reference_pole: complex
        Cartesian coordinates of pole.
    tau: float
        First moment of the final filter.

    Returns
    -------
    np.ndarray
        The transfer function with cols [f_axis, H_dosc(f)],
        (n, 2), dtype=complex
    """

    # aliases
    p = reference_pole
    pc = p.conjugate()

    # make transfer-function panel
    npts = np.arange(f_start, f_end, df).shape[0]
    H_dosc = np.ndarray([npts, 2], dtype=complex)

    # setup f_axis
    H_dosc[:, 0] = np.arange(f_start, f_end, df)
    f_axis = H_dosc[:, 0]

    # compute the complex transfer function along the /jw/ axis
    jw = 1j * 2.0 * np.pi * f_axis

    H_dosc[:, 1] = p * pc / ((tau * jw + p) * (tau * jw + pc))

    # return
    return H_dosc


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_phase_spectrum(
    f_start: float, f_end: float, df: float, reference_pole: complex, tau: float
) -> np.ndarray:
    r"""
    Computes the phase of H_dosc(f) from the analytic form

                                   / w + wn sin(th) \
        angle_H_dosc(w) = - arctan|------------------|
                                   \   wn cos(th)   /

                                     / w - wn sin(th) \
                            - arctan|------------------|
                                     \  wn cos(th)    /

    Temporal scaling is applied with the substitution w -> tau w.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    reference_pole: complex
        Cartesian coordinates of pole.
    tau: float
        First moment of the final filter.

    Returns
    -------
    np.ndarray
        The phase spectrum with cols [f_axis, angle_H_dosc(f)],
        (n, 2), dtype=complex
    """

    # make phase-spectrum panel
    npts = np.arange(f_start, f_end, df).shape[0]
    angle_H_dosc = np.ndarray([npts, 2], dtype=float)

    # setup f_axis
    angle_H_dosc[:, 0] = np.arange(f_start, f_end, df)
    tau_f = tau * angle_H_dosc[:, 0]

    # from /reference_pole/ extract wn and theta
    wns, thetas = design.extract_wn_and_theta(np.array([reference_pole]))
    fn = wns[0] / (2.0 * np.pi)
    theta = thetas[0]

    # compute the phase
    angle_H_dosc[:, 1] = -np.arctan2(
        tau_f + fn * np.sin(theta), fn * np.cos(theta)
    ) - np.arctan2(tau_f - fn * np.sin(theta), fn * np.cos(theta))

    return angle_H_dosc


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_spectra(
    f_start: float, f_end: float, df: float, reference_pole: complex, tau: float
) -> np.ndarray:
    """Generate gain, phase and group-delay spectra.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    reference_pole: complex
        Cartesian coordinates of pole.
    tau: float
        First moment of the final filter.

    Returns
    -------
    np.ndarray
        Spectra panel [n_pts, 4] with cols [f, g(f), phase(f), grpdly(f)].
    """

    # fetch the transfer function for gain calculation
    H: np.ndarray = generate_transfer_function(
        f_start, f_end, df, reference_pole, tau
    )

    # fetch the phase spectrum for group-delay calculation
    phase_H: np.ndarray = generate_phase_spectrum(
        f_start, f_end, df, reference_pole, tau
    )

    # calculate and return spectra
    return transfer_function_tools.calculate_spectra(H, phase_H)



"""
-------------------------------------------------------------------------------

Common tools for digital designs.

-------------------------------------------------------------------------------
"""

import numpy as np
from enum import Enum
from types import ModuleType

from numpy.polynomial import Polynomial

from irides.tools import impulse_response_tools as ir_tools
from irides.tools import analog_to_digital_conversion_tools as a2d_tools
from irides.resources.core_enumerations import FilterDesignType
from irides.tools.design_tools import StageTypes
from irides.resources.containers.discrete_sequence import DiscreteSequence
from irides.resources.containers.transfer_functions import (
    IIRTransferFunctionDiscreteTime,
)


class FrequencyBand(Enum):
    """Defines the set feasible frequency bands, whether low or high"""

    LOW = 0
    HIGH = 1


# noinspection PyPep8Naming
def construct_transfer_function(
    mu: float,
    filter_order: int,
    design_module: ModuleType,
    filter_class: FilterDesignType = FilterDesignType.LEVEL,
    frequency_band: FrequencyBand = FrequencyBand.LOW,
    T: float = 1,
) -> IIRTransferFunctionDiscreteTime:
    """Constructs H(z) from H(s) using impulse-invariant design.

    The transfer-function representation is multistage. Accordingly, the
    stages are built but not combined; instead the per-stage polynomials
    are passed to the constructor of the TransferFunction object for storage
    and evaluation.

    Parameters
    ----------
    mu: float
        Temporal scale
    filter_order: int
        The order of the filter
    design_module: ModuleType
        A design module that implements .designs(filter_order)
    filter_class: FilterDesignType
        Class of filter {level|slope|curve}
    frequency_band: FrequencyBand
        Specifies LOW or HIGH frequency (default = LOW)
    T: float
        Sampling rate (default = 1)

    Returns
    -------
    IIRTransferFunctionDiscreteTime
        Object that contains a transfer-function representation
    """

    # validate order
    ir_tools.validate_filter_order_or_die(design_module, filter_order)

    # fetch analog design
    filter_design = design_module.designs(filter_order)

    # prepare zplane poles
    zplane_poles_lf = a2d_tools.zplane_poles_from_analog_design(
        design_module, mu, filter_order, T
    )
    zplane_poles_signed = convert_frequency_band(
        zplane_poles_lf, frequency_band
    )

    # build the polynomial fragments stage by stage
    num_polys = np.array([])
    den_polys = np.array([])

    for stage in filter_design["stages"]:
        if stage["type"] == StageTypes.EMA.value:
            # poly: 1 - p z^-1
            zpole = np.real(zplane_poles_signed[stage["indices"]])[0]
            p = Polynomial([1.0, -zpole])

            zpole_lf = np.real(zplane_poles_lf[stage["indices"]])[0]
            p_gain_adj = Polynomial([1.0, -zpole_lf])

        else:
            # poly: 1 - (p + p*) z^-1 + pp* z^-2
            zpoles = zplane_poles_signed[stage["indices"]]
            a0 = 1.0
            a1 = -np.real(np.sum(zpoles))
            a2 = np.real(np.prod(zpoles))
            p = Polynomial([a0, a1, a2])

            zpoles_lf = zplane_poles_lf[stage["indices"]]
            a0 = 1.0
            a1 = -np.real(np.sum(zpoles_lf))
            a2 = np.real(np.prod(zpoles_lf))
            p_gain_adj = Polynomial([a0, a1, a2])

        # append polynomial fragment to frag array
        num_polys = np.append(num_polys, Polynomial(p_gain_adj(1.0)))
        den_polys = np.append(den_polys, p)

    # construct container and return
    return IIRTransferFunctionDiscreteTime(
        filter_order, num_polys, den_polys, filter_class
    )


def compute_sample_dtft_spectra(
    ds_in: DiscreteSequence,
    freq_max: float,
    n_freqs: int,
    v_col: int = 0,
) -> DiscreteSequence:
    """Implements the DTFT analysis equation for an input sample time series.

    X(e^jW) is returned by this function, using the equation

        X(e^jW) = sum_{i=0}^m x[n] e^jWn.

    Parameters
    ----------
    ds_in: DiscreteSequence
        Input sample time sequence
    freq_max: float
        Maximum frequency (eg 0.5)
    n_freqs: int
        Number of frequencies in the [0, pi) range
    v_col: int
        The column in ds_in.v_axes[:, `v_col`] to use (default = 0)

    Returns
    -------
    DiscreteSequence
        DTFT analysis output, n_axis -> frequencies, v_axis -> complex spectra
    """

    freqs = np.linspace(0, freq_max, n_freqs)
    coef = 1j * 2 * np.pi
    ds_out = DiscreteSequence(0, n_freqs, v_type="complex")

    for i, freq in enumerate(freqs):
        ejWn = np.exp(coef * freq * ds_in.n_axis)
        ds_out.v_axis[i] = np.sum(ds_in.v_axes[:, v_col] * ejWn)

    return ds_out


def identify_frequency_bands_of_zplane_roots(zplane_roots: np.ndarray) -> list:
    """Returns list of {LOW|HIGH} frequency types by inspecting zplane roots

    The frequency band is LOW if Re(root) >= 0, HIGH otherwise.

    Parameters
    ----------
    zplane_roots: np.ndarray
        Coordinates of zplane roots (float or complex)

    Returns
    -------
    list
        List of FrequencyBand values
    """

    selector = lambda b: FrequencyBand.LOW if b else FrequencyBand.HIGH
    is_nonnegative_real = np.real(zplane_roots) >= 0.0

    return list(map(selector, is_nonnegative_real))


def convert_frequency_band(
    zplane_roots: np.ndarray, target_frequency_band: FrequencyBand
) -> np.ndarray:
    """Converts `zplane_roots` to low- or high-freq roots given target type

    Parameters
    ----------
    zplane_roots: np.ndarray
        Coordinates of zplane roots (float or complex)
    target_frequency_band: FrequencyBand
        Target band, whether LOW or HIGH

    Returns
    -------
    np.ndarray
        Updated zplane_roots such that all roots have the specified freq band
    """

    # list of actual frequency bands
    freq_bands = identify_frequency_bands_of_zplane_roots(zplane_roots)

    # make list of signs so that (sign * zplane_pole) is in the spec'd freq band
    sign_converter = (
        lambda freq_type: 1.0 if freq_type == target_frequency_band else -1.0
    )

    signs_to_apply = np.array(list(map(sign_converter, freq_bands)))

    # apply and return
    return signs_to_apply * zplane_roots


# noinspection SpellCheckingInspection
def qz_coefs_from_zplane_poles(zplane_poles: np.ndarray) -> np.ndarray:
    """Returns array of Q(z) coefficients given an input array of z-poles

    Expressed in terms of zeta ( = z^-1 ), the zplane poles associate
    with simple polynomials according to

        (1 - p_1 zeta) (1 - p_2 zeta) ....

    Once multiplied through, the polynomial reads

        1 + a_1 zeta + a_2 zeta^2 + ....

    This routine returns the array [1, a_1, a_2, ...].

    Parameters
    ----------
    zplane_poles: np.ndarray
        Array of zplane poles (eg from zplane_poles_from_analog_design)

    Returns
    -------
    np.ndarray
        Coefficients to Q(z) polynomial
    """

    # Note that .fromroots() treats the input array as associated with
    #   (x - r[0]) (x - r[1]) ... (x - r[n-1])
    # rather than the simple polynomials expressed above. So, the
    # result from the routine needs to be reversed in order to correspond.

    return np.real(Polynomial.fromroots(zplane_poles).coef[::-1])


def build_a_coefs_matrix_from_qz_coefs(
    design_analog_level,
    mu: float,
    order: int,
) -> np.ndarray:
    """Builds a matrix of Q(z) coefs for subsequent use.

    The (order + 1) coefs from Q(z) are [a_0, a_1, a_2, ...]. ZIR analysis
    requires a matrix of the form

            | a_1  a_2  a_3 |
        A = | a_2  a_3  0   |
            | a_3  0    0   |

    which this function constructs.

    Parameters
    ----------
    design_analog_level:
        an analog design module for level filters
    mu: float:
        timescale of the digital filter
    order: int:
        order of the filter

    Returns
    -------
    np.ndarray:
        An A matrix
    """

    # fetch the Q(z) coefs
    qz_coefs = qz_coefs_from_zplane_poles(
        a2d_tools.zplane_poles_from_analog_design(
            design_analog_level, mu, order
        )
    )

    # remove qz_coefs[0] term, rename to a_coefs
    a_coefs = qz_coefs[1:]
    a_coefs_padded = np.pad(a_coefs, (0, order + 1))

    # build matrix
    A = np.zeros((order, order))
    for i in np.arange(0, order):
        A[i, :] = np.roll(a_coefs_padded, -i)[:order]

    # return
    return A


def zir_coefs_at_fixed_horizon(
    design_analog_level,
    design_digital_level,
    filter_signature_digital_level,
    mu: float,
    order: int,
) -> np.ndarray:
    """Coefficients for a fixed-horizon ZIR filter.

    The ZIR filter in this case appears in the

        Y_zir(z) = H_zir(z) Y(z)

    system equation whose inverse transform is

        y_zir[n; mu] = h_zir[n; mu] .conv. y[n].

    Use the H_zir(z) filter when propagating forward in `n` while
    maintaining a fixed forward horizon of `n + mu`.

    The form of H(z) in the context of Y(z) = H(z) X(z), the
    original filter, the pema and bessel H(z) structure is

                              gain_adj
        H(z) = ------------------------------------------  .
                1 + a_1 z^-1 + a_2 z^-2 + ... + a_m z^-m

    Using this form and simplifying to a 3rd-order system, the H_zir(z)
    transfer function reads

        H_zir(z) = gain_adj^-1 x (
            - (a_1 + a_2 + a_3) z^-1 - (a_2 + a_3) z^-2 - a_3 z^-3
        ).

    For Y_zir(z) = H_zir(z) H(z) X(z), the fixed-horizon ZIR impulse response is

        ~~ validate -- I think there's delta[n - k] in each item
        h_zir[n; mu] = - gain_adj^-1 x (
            (a_1 + a_2 + a_3) h[mu-1] + (a_2 + a_3) h[mu-2] + a_3 h[mu-3]
        )
        ~~

    Parameters
    ----------
    design_analog_level:
        an analog design module for level filters
    design_digital_level:
        a digital design module for level filters
    filter_signature_digital_level:
        a filter-signature module for level filters
    mu: float
        timescale of the digital filter
    order: int
        order of the filter

    Returns
    -------
    np.ndarray:
        array of zir coefs
    """

    # coefs of Q(z)
    qz_coefs = qz_coefs_from_zplane_poles(
        a2d_tools.zplane_poles_from_analog_design(
            design_analog_level, mu, order
        )
    )

    # gain_adj
    gain_adj = design_digital_level.gain_adjustment(mu, order)

    # hl[n] values at n = mu and its lags
    hl_weights = filter_signature_digital_level.generate_impulse_response(
        0, 4 * mu, mu, order
    ).v_axis[mu - order : mu + 1]

    # calculate zir coefs
    zir_coef_vals = np.zeros_like(qz_coefs)

    # these equations need to be checked (probably ok but I don't understand
    # how the math aligns with the code here)
    for i in np.arange(1, zir_coef_vals.shape[0]):
        alpha_ic_coefs = -qz_coefs[i:]
        hn_coefs = hl_weights[-(order + 1) + i :][::-1]
        zir_coef_vals[i] = (alpha_ic_coefs @ hn_coefs) / gain_adj

    return zir_coef_vals


def hzir_for_variable_horizon(
    yn_stream: np.ndarray,
    design_analog_level,
    design_digital_level,
    mu: float,
    order: int,
) -> DiscreteSequence:
    """Impulse response for the ZIR continuation of a yn stream.

    The returned impulse response h_zir[ell] is the IIR component of a
    continuation of a filtered stream y[n] = (h * x)[x], where h[n] may be
    a level or slope filter. That is, at sample `n`, further input x[n] is
    frozen to _x_[n], after which the ZIR response of the filter h[n] continues
    due to the energy stored in the filter.

    The real-world sample axis is denoted `n` while the continuation axis
    on which the ZIR response propagates is denoted by `ell` (after the
    Cyrillic character).

    The complete ZIR continuation for a level filter is

        y_zir[ell] = (h_zir * h)[ell] + _x_[n] (h * u)[ell].

    The impulse response of h_zir[ell] is `order` long in length, and the
    term-by-term entries go as

        h_zir[k] = - 1 / gain_adj alpha_ic[k],  k = 0, 1, 2

    where

                   | a_1  a_2  a_3 |   | y[n-1] |
        alpha_ic = | a_2  a_3  0   |   | y[n-2] |
                   | a_3  0    0   |   | y[n-3] |

    and y[n-k] are the input initial conditions.


    Parameters
    ----------
    yn_stream: np.ndarray
        Stream of output values from a preceding filter. The stream is to be
        order + 1 in length and ordered in ascending time.
    design_analog_level:
        an analog design module for level filters
    design_digital_level:
        a digital design module for level filters
    mu: float
        timescale of the digital filter
    order: int
        order of the filter

    Returns
    -------
    DiscreteSequence:
        The sequence of h_zir `order` samples long.
    """

    # build A matrix from coefs of Q(z)
    A = build_a_coefs_matrix_from_qz_coefs(design_analog_level, mu, order)

    # convert yn_stream to yn initial conditions
    yn_ic = yn_stream[1 : order + 1][::-1]

    # build alpha coefs based on A @ yn_ic
    alpha_ic_coefs = A @ yn_ic

    # gain_adj
    gain_adj = design_digital_level.gain_adjustment(mu, order)

    # prepare an impulse response
    ds_hzir = DiscreteSequence(0, order)
    ds_hzir.v_axis = -alpha_ic_coefs / gain_adj

    return ds_hzir


def hzir_for_fixed_horizon(
    hn_level: DiscreteSequence,
    design_analog_level,
    design_digital_level,
    mu: float,
    order: int,
    ell_select: int = None,
) -> DiscreteSequence:
    """Impulse response for the fixed-horizon.

    The impulse response goes as

        h_zir[n] = - 1 / gain_adj x
            (
                beta_1 delta[n-1] + beta_2 delta[n-2] + beta_3 delta[n-3] + ...
            )

    where

               | a_1  a_2  a_3 |   | hl[ell]   |
        beta = | a_2  a_3  0   |   | hl[ell-1] | .
               | a_3  0    0   |   | hl[ell-2] |

    The z-transform then reads

        H_zir(z) = - 1 / gain_adj x
            (
                beta_1 z^{-1} + beta_2 z^{-2} + beta_3 z^{-3} + ...
            ).

    Parameters
    ----------
    design_analog_level
    design_digital_level
    mu
    order

    Returns
    -------
    DiscreteSequence:
        The sequence of hzir `order` samples long.

    """

    # build A matrix from coefs of Q(z)
    A = build_a_coefs_matrix_from_qz_coefs(design_analog_level, mu, order)

    # pick hl values from hn_level (note reverse order)
    ell = mu if ell_select is None else ell_select
    hl = hn_level.v_axis[ell - order + 1 : ell + 1][::-1]

    # compute beta coefs
    beta = A @ hl

    # gain_adj
    gain_adj = design_digital_level.gain_adjustment(mu, order)

    # prepare the hl_zir impulse response
    # observe that beta_0 = 0
    ds_hzir = DiscreteSequence(0, order + 1)
    ds_hzir.v_axis[1:] = -beta / gain_adj

    # return
    return ds_hzir


# noinspection SpellCheckingInspection
def upsample_sequence(
    original_sequence: DiscreteSequence, upsample_stride: int
) -> DiscreteSequence:
    """Returns an upsampled sequence based on the input sequence.

    Parameters
    ----------
    original_sequence: DiscreteSequence
        Sequence to be upsampled
    upsample_stride: int
        Stride of adjacent upsampled values (lefthand point plus intermediate zeros)

    Returns
    -------
    DiscreteSequence
        Upsampled sequence
    """

    # set up new discrete sequence
    len_orig = original_sequence.len
    len_up = upsample_stride * len_orig
    n_start = original_sequence.n_axis[0]
    n_cols = original_sequence.v_axes.shape[1]
    ds_up = DiscreteSequence(n_start, n_start + len_up, n_cols)

    # for each v_axes column, expand
    for i in range(n_cols):
        # original values
        v_axis = original_sequence.v_axes[:, i][:, np.newaxis]

        # block of zeros where column count = (stride - 1)
        zeros_block = np.zeros((len_orig, upsample_stride - 1))

        # augment and collapse into a single vector
        ds_up.v_axes[:, i] = np.ravel(np.hstack((v_axis, zeros_block)))

    # return
    return ds_up



"""
-------------------------------------------------------------------------------

Analog poly-ema filter design:

Captures the design details necessary to design a polyema level filter,
and also captures details of its characteristic behavior.

The level filter is a low-pass filter that can be used itself or as
a component for a slope, curvature or mirror filter.

Wavenumbers are calculated numerically in Python in `calculate_wavenumber.py`.

-------------------------------------------------------------------------------
"""

import numpy as np
import scipy.special
import sys

from irides.resources.containers import points
from irides.tools import design_tools
from irides.tools.design_tools import StageTypes


# noinspection SpellCheckingInspection
def design_id() -> str:
    """
    Since `design` modules are passed around as objects in this framework,
    a `design_id` gives a unique name.

    Returns
    -------
    str
        The id of this design.
    """

    return "polyema-level"


def design_type() -> design_tools.FilterDesignType:
    """This is a level filter"""

    return design_tools.FilterDesignType.LEVEL


# noinspection SpellCheckingInspection
def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns valid filter orders for this filter design.

    Parameters
    ----------
    strict: bool
        When true, only orders where h(0)=0 are included.

    Returns
    -------
    np.ndarray
        Array of valid filter orders
    """

    return design_tools.get_valid_filter_orders(
        design_tools.FilterDesignType.LEVEL, strict
    )


"""
-------------------------------------------------------------------------------

Filter-specific definitions

-------------------------------------------------------------------------------
"""


# noinspection SpellCheckingInspection
def designs(filter_order: int) -> dict:
    """Design parameters as a function of filter order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    dict
        Design dictionary.
    """

    # design given the order of the filter
    m = filter_order
    design = {
        "stages": [
            {"type": StageTypes.EMA.value, "indices": np.array([i])}
            for i in range(m)
        ],
        "poles": -m * np.ones(m, dtype=complex),
        "poly_coeffs": [scipy.special.binom(m, k) for k in range(m + 1)],
    }

    # return the design
    return design


def integer_sequence_a055897(n: int) -> int:
    """
    Integer sequence related peak impulse-response value.

    See https://oeis.org/A055897

    Parameters
    ----------
    n: int
        Sequence index.

    Returns
    -------
    int
        Sequence value.
    """

    return n * np.power((n - 1), n - 1)


# noinspection SpellCheckingInspection
def convert_reference_poles_to_temporal_scales(
    reference_poles: np.ndarray,
) -> np.ndarray:
    """Convert a single pole to the equivalent temporal scale.

    The relation between a reference pole and temporal scale is

        tau = 1 / abs(reference_pole)

    It is expected that imag(reference_pole) = 0 (since the scope of this
    function is in the polyema design, rather than damped oscillator design).
    Nonetheless, it is canonical to calculate the radius of the pole in the
    s-plane.

    Parameters
    ----------
    reference_poles: np.ndarray (dtype = complex)
        Reference poles to convert to temporal scales.

    Returns
    -------
    np.ndarray
        tau-values associated with reference poles.

    """

    return 1.0 / np.abs(reference_poles)


def tau_per_stage(tau: float, filter_order: int) -> float:
    """
    Calculates the temporal scale factor /tau_m/ for any of the identical
    ema stages along the polyema cascade:

        tau_per_stage_value = tau / filter_order .

    The cascade of /filter_order/ identical ema stages generates a combined
    first moment of /tau/.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The first moment for each (identical) stage: `tau_per_stage_value`.
    """

    # this is a simple calculation
    tau_per_stage_value = tau / filter_order

    # return
    return tau_per_stage_value


"""
-------------------------------------------------------------------------------

Standard features

-------------------------------------------------------------------------------
"""


def unit_peak_time(filter_order: int) -> float:
    """
    As per doc string for peak_value_coordinate(..),

        t* = (m - 1) / m  tau .

    This function returns the filter-order-based coefficient

        (m - 1) / m

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The timepoint where h(t) peaks, for unit delay.
    """

    # alias
    m = filter_order

    # return
    return (m - 1.0) / m


def unit_peak_value(filter_order: int) -> float:
    """
    As per doc string for peak_value_coordinate(..),

                        m (m-1)^(m-1)
        h(t*) * tau = -----------------
                       Gamma(m) e^(m-1)

    Function returns the right-hand side.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The value of h(t_peak), for unit delay.

    """

    # alias
    m = filter_order

    # construct fraction
    numerator = integer_sequence_a055897(m)
    denominator = scipy.special.gamma(m) * np.exp(m - 1)

    # return
    return numerator / denominator


# noinspection PyUnusedLocal
def peak_value_coordinate(tau: float, filter_order: int, dt=0.0):
    """
    Low-pass polyema filters have the characteristic behavior of
    having a single maximum value along its impulse response h(t).

                  ....x...  <---- h(t*)
              ....        ....
           ...                ..
         ..                     ...
        .                          ....
    ----|-------------|-------------------------> t
        0
                     t*

    From Mathematica script polyema_inline_slope.nb, the
    time t* where h(t) attains is maximum and the value of h(t*) are

              m-1                     m (m-1)^(m-1)        1
        t* = ----- tau,  and h(t*) = --------------- -------------,
               m                        Gamma(m)      e^(m-1) tau

    where m = /filter_order/. Note that
    the sequence m (m-1)^(m-1) is identified as https://oeis.org/A055897.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.
    dt: float
        Unused.

    Returns
    -------
    ContinuousTimePoint
        Object `peak_coord` with .time and .value fields.
    """

    # alias
    m = filter_order

    # put together into a point coordinate
    peak_coord = points.ContinuousTimePoint(
        unit_peak_time(m) * tau, unit_peak_value(m) / tau
    )

    # return
    return peak_coord


# noinspection SpellCheckingInspection,PyUnusedLocal
def moment_value_generator(moment: int, tau: float = 1.0, dt: float = 0.0):
    """
    Returns an anonymous function with the single argument (order: int),
    where order is a `filter_order`. The returned function captures the `tau`
    and `dt` arguments in this calling function.

    Note that the generated functions are not purely continuous time.
    Instead, they bridge continuous and discrete time by applying order-dt
    corrections to the moments. This is required because the CT series are
    in fact implemented in DT.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # generator defs with dt capture
    # noinspection PyUnusedLocal
    def m0_gen(order: int) -> float:
        if order == 1:
            return 1.0 + (dt / tau) / 2.0
        else:
            return 1.0

    # noinspection PyUnusedLocal
    def m1_gen(order: int) -> float:
        return tau

    # noinspection PyUnusedLocal
    def m2_gen(order: int) -> float:
        return ((order + 1.0) / order) * np.power(tau, 2)

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection,PyPep8Naming
def autocorrelation_peak_and_stride_values(
    tau: float, filter_order: int
) -> dict:
    """
    The zero-lag S-ACF parameter, kh(0), is analytic:

                       Gamma(m-1/2)
        kh(0) = ----------------------------- ,
                 2 sqrt(pi) (tau/m) Gamma(m)

    where m = filter_order.

    The stride xi_5pct where sacf(lag=xi_5pct) = 0.05 is reported numerically
    (based on results from Mathematica).

    The zero-lag kh(0) scales as 1 / tau, and the stride xi_5pct scales as tau.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Peak value (scaled by 1/tau) and 5% pct stride (scaled by tau)
    """

    def kh0_gen(order: int) -> float:
        """Implements equation for kh(0) value, tau=1"""

        gamma = scipy.special.gamma
        return gamma(order - 0.5) / (
            2.0 * np.sqrt(np.pi) * (1.0 / order) * gamma(order)
        )

    # see Mathematica polyema_level.nb
    unit_sacf_values = {
        1: {"kh0": kh0_gen(1), "xi_5pct": 2.99573},
        2: {"kh0": kh0_gen(2), "xi_5pct": 2.37193},
        3: {"kh0": kh0_gen(3), "xi_5pct": 1.97288},
        4: {"kh0": kh0_gen(4), "xi_5pct": 1.71918},
        5: {"kh0": kh0_gen(5), "xi_5pct": 1.54189},
        6: {"kh0": kh0_gen(6), "xi_5pct": 1.40953},
        7: {"kh0": kh0_gen(7), "xi_5pct": 1.30603},
        8: {"kh0": kh0_gen(8), "xi_5pct": 1.22228},
    }
    zero_values = {"kh0": 0.0, "xi_5pct": 0.0}

    # check and return
    if filter_order in unit_sacf_values.keys():
        ans = unit_sacf_values[filter_order]
        ans["kh0"] /= tau
        ans["xi_5pct"] *= tau
    else:
        ans = zero_values

    return ans


# noinspection SpellCheckingInspection,PyPep8Naming
def scale_correlation_stride_and_residual_values(
    tau: float, filter_order: int
) -> dict:
    """Scale decorrelation strides and residual correlations.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Scale decorrelation (scales as `tau`) and residual correlation
        (independent of `tau`).
    """

    # values from Mathematica
    unit_sscf_values = {
        1: {"sti_5pct": 1598.00, "residual": 0.05},
        2: {"sti_5pct": 27.4358, "residual": 0.05},
        3: {"sti_5pct": 11.1683, "residual": 0.05},
        4: {"sti_5pct": 7.27680, "residual": 0.05},
        5: {"sti_5pct": 5.60510, "residual": 0.05},
        6: {"sti_5pct": 4.68270, "residual": 0.05},
        7: {"sti_5pct": 4.09790, "residual": 0.05},
        8: {"sti_5pct": 3.69310, "residual": 0.05},
    }

    # fetch values for order and scale
    sscf_values = unit_sscf_values[filter_order]
    sscf_values["sti_5pct"] *= tau

    # return
    return sscf_values


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def full_width_generator(tau: float = 1.0, dt: float = 0.0):
    """
    Returns anonymous function for the full-width. `tau` and `dt` are captured
    here, and the anonymous function has one argument, (order: int), the filter
    order.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    def fw_gen(order: int):
        return 2.0 * np.sqrt(1.0 / order) * tau

    return fw_gen


# noinspection PyUnusedLocal
def gain_at_dc(tau: float) -> float:
    """Filter gain at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return 1.0


# noinspection PyUnusedLocal
def phase_at_dc(tau: float) -> float:
    """Filter phase at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return 0.0


def group_delay_at_dc(tau: float) -> float:
    """Filter group delay at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float
        The gain at DC frequency.
    """

    return tau


def cutoff_frequency(tau: float, filter_order: int) -> float:
    """
    The cutoff frequency /wc/ is the radial frequency wc where

        |H(wc)|^2 = 1/2 .

    That frequency is

               m
        wc = ----- sqrt( 2^(1/m) - 1 ) .
              tau

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The cutoff frequency.
    """

    # alias
    m = float(filter_order)

    # calculate the cutoff frequency
    wc = m / tau * np.sqrt(np.power(2.0, 1.0 / m) - 1.0)

    # return
    return wc


# noinspection PyUnusedLocal
def gain_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The cutoff gain is always sqrt(1/2). The purpose of this function is
    to adhere to the phase- and group-delay-cutoff api so that any of the
    gain/phase/group-delay values @ cutoff can be determined uniformly.
    """

    return np.sqrt(1.0 / 2.0)


# noinspection PyUnusedLocal
def phase_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The phase at cutoff frequency `wc`. The phase is independent of
    temporal scale `tau`, and the analytic expression is

        phase(wc) = -m arctan( sqrt( 2^(1/m) - 1 ) ) .

    Parameters
    ----------
    tau: float
        Unused but present for a consistent api
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The phase at the cutoff frequency.
    """

    # calculate from analytic expression
    m = filter_order
    phase_wc = -m * np.arctan(np.sqrt(np.power(2.0, 1.0 / m) - 1.0))

    # return
    return phase_wc


def group_delay_at_cutoff(tau: float, filter_order: int) -> float:
    """
    The group delay at cutoff frequency /wc/. The group delay here is

        grp-delay(wc) = 2^(-1/m) tau .

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The group delay at cutoff.
    """

    # calculate the cutoff frequency
    grp_delay = np.power(2, -1.0 / filter_order) * tau

    # return
    return grp_delay


# noinspection SpellCheckingInspection
def wireframe(tau: float, filter_order: int):
    """Wireframe timepoint for this level filter.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter (unused)

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    return design_tools.generate_level_wireframe_from_spectra(
        sys.modules[__name__], tau, filter_order
    )


# noinspection SpellCheckingInspection
def wavenumber(filter_order: int) -> float:
    """Report the wavenumber for this filter as a function of order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The wavenumber.
    """

    wavenumbers = {
        1: 0.64780859,
        2: 0.72623682,
        3: 0.74588861,
        4: 0.75626296,
        5: 0.76143075,
        6: 0.76197976,
        7: 0.75833512,
        8: 0.75110948,
    }

    return wavenumbers[filter_order]


# noinspection SpellCheckingInspection
def uncertainty_product(filter_order: int) -> dict:
    """Dict of TW, BW, and UCP for this filter order.

    Parameters
    ----------
    filter_order: int
        The order of the filter

    Returns
    -------
    dict
        TW is temporal width, BW is spectral width, and UCP is the
        uncertainty product.
    """

    uncertainty_products = {
        1: {"TW": 0.5, "BW": np.inf, "UCP": np.inf},
        2: {"TW": 0.433013, "BW": 2.0, "UCP": 0.866025},
        3: {"TW": 0.372678, "BW": 1.73205, "UCP": 0.645497},
        4: {"TW": 0.330719, "BW": 1.78885, "UCP": 0.591608},
        5: {"TW": 0.3, "BW": 1.88982, "UCP": 0.566947},
        6: {"TW": 0.276385, "BW": 2.0, "UCP": 0.552771},
        7: {"TW": 0.257539, "BW": 2.11058, "UCP": 0.543557},
        8: {"TW": 0.242061, "BW": 2.2188, "UCP": 0.537086},
    }

    return uncertainty_products[filter_order]



"""
-------------------------------------------------------------------------------

Digital poly-ema level-filter design.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.analog import polyema_level as dsgn_alg_pema
from irides.tools import analog_to_digital_conversion_tools as a2d_tools
from irides.tools import design_tools


def design_id() -> str:
    """Design id of this filter."""

    return "digital-{0}".format(dsgn_alg_pema.design_id())


def design_type() -> design_tools.FilterDesignType:
    """This is a level filter"""

    return dsgn_alg_pema.design_type()


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns valid filter orders for this filter design."""

    return dsgn_alg_pema.get_valid_filter_orders(strict)


def convert_mu_to_pole(mu: float, filter_order: int) -> float:
    """Computes `p` value from parameters"""

    return mu / (filter_order + mu)


def minimum_mu_value(filter_order: int) -> float:
    """Returns the minimum mu value for a given filter order"""

    return filter_order


def gain_adjustment(mu: float, filter_order: int) -> float:
    """Returns the gain adjustment so that the filter's level gain is unity."""

    # compute the associated pole
    p = convert_mu_to_pole(mu, filter_order)

    # compute and return the gain adjustment
    return np.exp(filter_order * np.log(1 - p))


def initial_hn_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns analytically calculated values of h[0], h[1], and h[2].

    From a general single-pole product expansion of H(z), we find that

        h[0] = gain_adj
        h[1] = gain_adj * sum( digital_poles )
        h[2] = gain_adj * (sum( digital_poles )^2 - residual)

    where, as a 4th-order example to show the pattern,

        residual = p1 (p2 + p3 + p4) + p2 (p3 + p4) + p3 * p4.
    """

    # prerequisites
    zplane_poles = a2d_tools.zplane_poles_from_analog_design(
        dsgn_alg_pema, mu, filter_order
    )

    # calculate h[0]
    h0 = gain_adjustment(mu, filter_order)

    # calculate h[1]
    zplane_poles_sum = np.sum(zplane_poles)
    h1 = h0 * np.real(zplane_poles_sum)

    # calculate h[2]
    residual = np.real(
        np.sum(
            [
                zplane_poles[i] * np.sum(zplane_poles[i + 1 :])
                for i in range(filter_order - 1)
            ]
        )
    )
    h2 = h0 * np.real(np.power(zplane_poles_sum, 2) - residual)

    # pack and return as an array
    return np.array([h0, h1, h2])


# noinspection PyPep8Naming
def moment_values(mu: float, filter_order: int) -> np.ndarray:
    """Returns analytically calculated moment values (M0, M1, M2)"""

    M0 = 1.0
    M1 = mu
    M1_s = M1 / filter_order
    M2 = (1.0 + M1 + M1_s) * M1

    return np.array([M0, M1, M2])


def full_width_value(mu: float, filter_order: int) -> float:
    """Returns the analytically calculated FW value"""

    mu = mu
    m = filter_order
    return 2.0 * np.sqrt(mu * (m + mu) / m)


def analog_tau_value(mu: float, filter_order: int) -> float:
    """Returns `tau` associated with (`mu`, `filter_order`)"""

    # compute and return
    splane_poles = dsgn_alg_pema.designs(filter_order)["poles"]

    return a2d_tools.solve_for_tau_from_mu_and_splane_poles(mu, splane_poles)



"""
-------------------------------------------------------------------------------

Digital damped-oscillator filter-element design

-------------------------------------------------------------------------------
"""

import numpy as np
from typing import Union

from irides.design.analog import damped_oscillator as dsgn_alg_dosc
from irides.tools import analog_to_digital_conversion_tools as a2d_tools
from irides.tools import digital_design_tools as dd_tools


# -----------------------------------------------------------------------------
# Purely digital
# -----------------------------------------------------------------------------

# noinspection SpellCheckingInspection
def gain_adjustment(
    zplane_poles: np.ndarray, return_polar_coords=False
) -> Union[np.ndarray, tuple]:
    """Returns the gain adjustment for an array of zplane poles.

    Parameters
    ----------
    zplane_poles: np.ndarray
        The digital poles
    return_polar_coords: bool
        Returns (gain_adj, r, phi) as tuple of np.ndarrays if T (default F)

    Returns
    -------
    Union[np.ndarray, tuple]
        gain adjustment as an array, or tuple that includes polar coords
        of zplane_poles
    """

    # convert to polar form
    polar_poles = dsgn_alg_dosc.cast_to_polar_form(zplane_poles)
    r = polar_poles[:, 0]
    phi = polar_poles[:, 1]

    # generate polar form of low-frequency poles
    zplane_lf_poles = dd_tools.convert_frequency_band(
        zplane_poles, dd_tools.FrequencyBand.LOW
    )
    polar_lf_poles = dsgn_alg_dosc.cast_to_polar_form(zplane_lf_poles)
    phi_lf = polar_lf_poles[:, 1]

    # compute gain adjustment -- always at low frequency
    gain_adj = 1.0 - 2.0 * np.cos(phi_lf) * r + r * r

    if return_polar_coords:
        return gain_adj, r, phi
    else:
        return gain_adj


def initial_hn_values(zplane_poles: np.ndarray) -> np.ndarray:
    """Returns analytically calculated h[0], h[1], h[2] values.

    Parameters
    ----------
    zplane_poles: np.ndarray
        Poles in cartesian coordinates.

    Returns
    -------
    np.ndarray
        Array [len(zplane_poles), 3] of values h[0], h[1], h[2]
    """

    # prerequisites
    gain_adj, r, phi = gain_adjustment(zplane_poles, return_polar_coords=True)

    # calculate h[0]
    h0 = gain_adj

    # calculate h[1]
    h1 = 2.0 * np.cos(phi) * r * h0

    # calculate h[2]
    h2 = (1.0 + 2.0 * np.cos(2.0 * phi)) * r * r * h0

    # pack and return as an array
    return np.array([h0, h1, h2]).T


# noinspection SpellCheckingInspection,PyPep8Naming
def moment_values(zplane_poles: np.ndarray) -> np.ndarray:
    """Returns the requested moment for each z-plane pole value.

    Parameters
    ----------
    zplane_poles: np.ndarray
        z-plane poles in Cartesian coordinate

    Returns
    -------
    np.ndarray
        Moments [M0, M1, M2]
    """

    # prerequisites
    gain_adj, r, phi = gain_adjustment(zplane_poles, return_polar_coords=True)

    # 0th moment
    M0 = np.ones_like(r)

    # 1st moment -- M1 = H'(1)
    M1 = 2.0 * r * (np.cos(phi) - r) / gain_adj

    # 2nd moment -- M2 = H'(1) + H''(1)
    cos_phi = np.cos(phi)
    cos_2_phi = np.cos(2.0 * phi)

    #   compute H'
    H_p = 2.0 * r * (cos_phi - r) / gain_adj

    #   compute H''
    elmt = 1.0 + 2.0 * cos_2_phi - 6.0 * cos_phi * r + 3.0 * r * r
    H_pp = 2.0 * r * r * elmt / np.power(gain_adj, 2)

    #   compute M2
    M2 = H_pp + H_p

    return np.array([M0, M1, M2]).T


# noinspection SpellCheckingInspection
def full_width_value(zplane_poles: np.ndarray) -> np.ndarray:
    """Calculates the analytic full-width value(s), returns nan when not def'd.

    Parameters
    ----------
    zplane_poles: np.ndarray
        Poles in cartesian coordinates.

    Returns
    -------
    np.ndarray
        Full-width values
    """

    # prerequisites
    gain_adj, r, phi = gain_adjustment(zplane_poles, return_polar_coords=True)

    with np.errstate(invalid="ignore"):

        # compute
        arg = 2.0 * r * ((1.0 + r * r) * np.cos(phi) - 2.0 * r)
        # gain_adj = gain_adjustment(r, phi)

        # return
        return 2.0 * np.sqrt(arg) / gain_adj


# -----------------------------------------------------------------------------
# Mixed analog and digital
# -----------------------------------------------------------------------------

# noinspection SpellCheckingInspection
def tau_minimum_for_analog_pole(splane_poles: np.ndarray) -> np.ndarray:
    """Returns tau_min given an s-plane pole.

    tau_min is defined as the value of tau that leads to

        Re( zplane_pole ) = 1/2 .

    Parameters
    ----------
    splane_poles: np.ndarray
        Either pole of the complex-conjugate pair

    Returns
    -------
    np.ndarray
        minimum tau values
    """

    # by definition
    pdr = 0.5

    # compute and return
    return np.array(
        [
            a2d_tools.solve_for_tau_to_place_splane_pole_onto_zplane_real_axis(
                splane_pole, pdr
            )
            for splane_pole in splane_poles
        ]
    )



"""
-------------------------------------------------------------------------------

Digital Kiyasu--Thomson Bessel filter signature.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.analog import bessel_level as dsgn_a_l_bssl
from irides.filter_signatures.digital import polyema_level as fsig_d_l_pema
from irides.filter_signatures.digital import (
    damped_oscillator as fsig_d_dosc,
)

from irides.tools import impulse_response_tools as ir_tools
from irides.tools import analog_to_digital_conversion_tools as a2d_tools
from irides.resources.containers.discrete_sequence import DiscreteSequence
from irides.tools import digital_design_tools as dd_tools
from irides.tools.design_tools import StageTypes
from irides.tools.digital_design_tools import FrequencyBand


# noinspection SpellCheckingInspection
def generate_impulse_response(
    n_start: int,
    n_end: int,
    mu: float,
    filter_order: int,
    frequency_band: FrequencyBand = FrequencyBand.LOW,
    design=dsgn_a_l_bssl,
) -> DiscreteSequence:
    """Creates a discrete-time impulse response for the KT-Bessel filter.

    Parameters
    ----------
    n_start: int
        Start sample
    n_end: int
        End sample (exclusive)
    mu: float
        First moment of the impulse response
    filter_order: int
        Filter order
    frequency_band: FrequencyBand
        Specifies LOW or HIGH frequency (default = LOW)
    design:
        The name of the associated analog design module

    Returns
    -------
    DiscreteSequence
        Contains n-axis and h-axis
    """

    # validate
    ir_tools.validate_filter_order_or_die(design, filter_order)

    # fetch the analog design config for this filter order
    design_config = design.designs(filter_order)

    # aliases
    stages = design_config["stages"]
    splane_poles = design_config["poles"]

    # calc the number of stages
    n_stgs = len(design_config["stages"])

    # set up a discrete sequence to work with
    ds = DiscreteSequence(n_start, n_end, 1 + n_stgs)
    n_0plus_vec = ds.n_axis[ds.i_n_zero :]

    # correct analog-tau to achieve target digital-mu
    tau = a2d_tools.solve_for_tau_from_mu_and_splane_poles(mu, splane_poles)

    # create h[n]-generating fcxns
    # noinspection SpellCheckingInspection
    ema_filter_order = 1

    def hn_ema_gen(splane_poles_for_stage: np.ndarray) -> DiscreteSequence:
        pd = a2d_tools.convert_analog_poles_to_digital_poles(
            tau, splane_poles_for_stage
        )
        mu_stg = a2d_tools.first_moment_from_zplane_pole(pd)[0]
        return fsig_d_l_pema.generate_impulse_response(
            n_start, n_end, mu_stg, ema_filter_order, frequency_band
        )

    # noinspection SpellCheckingInspection
    def hn_dosc_gen(splane_poles_for_stage: np.ndarray) -> DiscreteSequence:
        pd = a2d_tools.convert_analog_poles_to_digital_poles(
            tau, splane_poles_for_stage
        )
        pd_signed = dd_tools.convert_frequency_band(pd, frequency_band)
        return fsig_d_dosc.generate_impulse_response(n_start, n_end, pd_signed)

    # pack into a dict
    hn_gen_dict = {
        StageTypes.EMA.value: hn_ema_gen,
        StageTypes.DOSC.value: hn_dosc_gen,
    }

    # iter across stages
    hn_cand = np.array([1.0])  # starting point for cascade
    for i, stage in enumerate(stages):
        # gen h[n] for this stage (dosc or ema)
        hn_stage = hn_gen_dict[stage["type"]](splane_poles[stage["indices"]])

        # convolve in new h[n] with running candidate
        hn_cand = np.convolve(hn_stage.v_axis, hn_cand)[: n_0plus_vec.shape[0]]

        # persist stage result (starting from second value column)
        ds.v_axes[ds.i_n_zero :, 1 + i] = hn_cand

    # persist the final cascade
    ds.v_axes[ds.i_n_zero :, 0] = hn_cand

    # return
    return ds



"""
-------------------------------------------------------------------------------

Digital multistage-box filter signature.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.design.digital import mbox_level as dsgn_d_l_mbox

from irides.tools import impulse_response_tools as ir_tools
from irides.resources.containers.discrete_sequence import DiscreteSequence


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_impulse_response(
    n_start: int, n_end: int, mu: float, filter_order: int
) -> DiscreteSequence:
    """Creates a discrete-time impulse response for the multistage-box filter.

    Note: n_end must be at least (2 mu + filter_order) in length in order to
          not clip the mbox impulse response.

    Parameters
    ----------
    n_start: int
        Start sample
    n_end: int
        End sample (exclusive)
    mu: float
        First moment of the impulse response
    filter_order: int
        Filter order

    Returns
    -------
    DiscreteSequence
        Contains n-axis and h-axis
    """

    # validate
    ir_tools.validate_filter_order_or_die(dsgn_d_l_mbox, filter_order)

    # alias
    m = filter_order

    # compute N_stage, which uses an adjusted `mu` value
    N_stage = dsgn_d_l_mbox.convert_adjusted_mu_to_N_stage(mu, filter_order)
    N_int = filter_order * N_stage

    # setup a discrete sequence
    ds = DiscreteSequence(n_start, n_end)

    # prepare for and exec m-fold box cascade
    h_box = np.ones(N_stage) / N_stage
    conv_mbox = np.array([1])
    for i in range(m):
        conv_mbox = np.convolve(h_box, conv_mbox)

    # trim box to length N
    conv_mbox = conv_mbox[:N_int]
    conv_mbox_neff_axis = np.arange(0, conv_mbox.shape[0])

    # identify overlapping indices of ds.n_axis and the implicit n-axis
    # of the conv_box array
    i_insertions = np.array(list(set(ds.n_axis) & set(conv_mbox_neff_axis)))

    # insert
    ds.v_axis[ds.i_n_zero : ds.i_n_zero + i_insertions.shape[0]] = conv_mbox[
        i_insertions
    ]

    # return
    return ds



"""
-------------------------------------------------------------------------------

Container for transfer functions as encoded by the numpy polynomial class

-------------------------------------------------------------------------------
"""

import numpy as np
from typing import Union

from irides.resources.core_enumerations import FilterDesignType


class IIRTransferFunctionDiscreteTime(object):
    """Container for IIR H(z) rational functions."""

    def __init__(
        self,
        filter_order: int,
        numerator_polynomials: np.ndarray,
        denominator_polynomials: np.ndarray,
        filter_class: FilterDesignType = FilterDesignType.LEVEL,
    ):
        """Initializes a Hz object.

        Instances of Polynomial objects that are passed in are to be expressed
        in terms of zeta = z^-1. The .value(z) method internally evaluates the
        polynomials at (1/z).

        Parameters
        ----------
        filter_order: int
            The filter order
        numerator_polynomials: np.ndarray
            Array of Polynomial objects
        denominator_polynomials: np.ndarray
            Array of Polynomial objects
        filter_class: FilterDesignType
            Class of filter {level|slope|curve}
        """

        # store locally
        self._filter_order = filter_order
        self._numerator_polynomials = numerator_polynomials
        self._denominator_polynomials = denominator_polynomials
        self._filter_class = filter_class

        # create capture
        exp = filter_class.value

        def eval_lead_coef_capture(
            zeta_value: Union[float, complex]
        ) -> Union[float, complex]:
            return (1 - zeta_value) ** exp

        self._eval_lead_coef = eval_lead_coef_capture

    def order(self) -> int:
        """Returns the order of H(z)"""

        return self._filter_order

    def value(
        self, z: Union[float, complex, np.ndarray]
    ) -> Union[float, complex, np.ndarray]:
        """Returns H(z) evaluated at point(s) z

        The method is

            H_impl(z) = (1 - z^(-1))^{0|1|2} x exp( log( H(z) ) ),

        which is significantly more numerically stable than just H(z).

        Parameters
        ----------
        z: Union[float, complex, np.ndarray]
            Value(s) of z to evaluate H(z).

        Returns
        -------
        Union[float, np.ndarray]
            Value(s) of H(z)
        """

        def single_value(zeta_value: Union[float, complex]) -> float:
            return self._eval_lead_coef(zeta_value) * np.exp(
                np.sum(
                    np.log([n(zeta_value) for n in self._numerator_polynomials])
                    - np.log(
                        [d(zeta_value) for d in self._denominator_polynomials]
                    )
                )
            )

        if np.isscalar(z):
            return single_value(1.0 / z)

        else:
            return np.array([single_value(1.0 / z_value) for z_value in z])

    def initial_value(self) -> float:
        """Returns value of h[0] computed from H(z)"""

        # calculate and return
        return self.value(np.inf)


# noinspection PyPep8Naming
class MboxTransferFunctionDiscreteTime(object):
    """Container for mbox H(z) transfer function."""

    def __init__(
        self,
        filter_order: int,
        N_stage: int,
        filter_class: FilterDesignType = FilterDesignType.LEVEL,
    ):
        """Initializes the Hz object.

        Parameters
        ----------
        filter_order: int
            The filter order
        N_stage: int
            The number of samples per stage (= N / m)
        filter_class: FilterDesignType
            Class of filter {level|slope|curve}
        """

        # store locally
        self._filter_order = filter_order
        self._n_stage = N_stage
        self._filter_class = filter_class

        # local calculations
        self._n_stage_is_even = (N_stage % 2) == 0
        self._log_gain_adj_per_stage = -np.log(N_stage)

        # create capture
        exp = filter_class.value

        def eval_lead_coef_capture(
            zeta_value: Union[float, complex]
        ) -> Union[float, complex]:
            return (1 - zeta_value) ** exp

        self._eval_lead_coef = eval_lead_coef_capture

    def order(self) -> int:
        """Returns the order of H(z)"""

        return self._filter_order

    def value(
        self, z: Union[float, complex, np.ndarray]
    ) -> Union[float, complex, np.ndarray]:
        """Returns H(z) evaluated at point(s) z.

        The method is

            H_impl(zeta) = (1 - zeta)^{0|1|2} x
                exp(
                    filter_order x (
                       log(gadj_per_stage)
                       + log(1 - zeta^{n_stage})
                       - log(1 - zeta)
                    )
                )

        which is fast and numerically stable. The degree of the lead (1 - zeta)
        polynomial depends on the filter class: level = 0, slope = 1, curve = 2.

        Parameters
        ----------
        z: Union[float, complex, np.ndarray]
            Value(s) of z to evaluate H(z).

        Returns
        -------
        Union[float, np.ndarray]
            Value(s) of H(z)
        """

        def single_value(zeta_value: Union[float, complex]) -> complex:
            if zeta_value == 1.0:
                v = self._eval_lead_coef(zeta_value) * np.exp(
                    self._filter_order
                    * (self._log_gain_adj_per_stage + np.log(self._n_stage))
                )
            elif zeta_value == -1.0 and self._n_stage_is_even:
                v = 0.0
            else:
                v = self._eval_lead_coef(zeta_value) * np.exp(
                    self._filter_order
                    * (
                        self._log_gain_adj_per_stage
                        + np.log(1 - zeta_value**self._n_stage)
                        - np.log(1 - zeta_value)
                    )
                )
            return v

        if np.isscalar(z):
            return single_value(1.0 / z)

        else:
            return np.array([single_value(1.0 / z_value) for z_value in z])

    def initial_value(self) -> float:
        """Returns value of h[0] computed from H(z)"""

        # calculate and return
        return self.value(np.inf)



"""
-------------------------------------------------------------------------------

Analog poly-ema composite-slope filter design:

Captures the design details for this filter. Design details are available
in Mathematica notebook polyema_composite_slope.nb. Wavenumbers are
calculated numerically in Python in `calculate_wavenumber.py`.

-------------------------------------------------------------------------------
"""

import numpy as np
import scipy.misc
import scipy.special
import sys

from irides.tools import design_tools


# noinspection SpellCheckingInspection
def design_id() -> str:
    """
    Since `design` modules are passed around as variables in this framework,
    a `design_id` gives a unique name.

    Returns
    -------
    str
        The id of this design.
    """

    return "polyema-composite-slope"


def design_type() -> design_tools.FilterDesignType:
    """This is a slope filter"""

    return design_tools.FilterDesignType.SLOPE


def get_minimum_valid_filter_order(strict: bool = False) -> int:
    """Returns minimum valid filter order."""

    return 1
    # return design_tools.get_minimum_valid_filter_order(design_type(), strict)


def get_valid_filter_orders(strict: bool = False) -> np.ndarray:
    """Returns array of valid filter orders based on strictness."""

    return np.array([1, 2, 3, 4, 5, 6, 7, 8])
    # return design_tools.get_valid_filter_orders(design_type(), strict)


# noinspection SpellCheckingInspection
def convert_armratio_order_params_to_tp_tm_params(
    arm_ratio: float, tau: float, filter_order: int
) -> tuple:
    """Converts the (alpha, m) tuple to a (tau+, tau-) tuple, scaled by tau."""

    # alias
    a = arm_ratio
    m = filter_order

    # calculate
    coef = m / (m + 1.0)
    tp = 2.0 * a / (1.0 + a) * coef * tau
    tm = 2.0 / (1.0 + a) * coef * tau

    return tp, tm


# noinspection SpellCheckingInspection
def convert_tp_tm_params_to_armratio_order_params(
    tp: float, tm: float, tau: float
) -> tuple:
    """Functionally inverts convert_armratio_order_params_to_tp_tm_params()."""

    # calculate
    arm_ratio = tp / tm
    tau_avg = (tp + tm) / (2.0 * tau)
    filter_order = tau_avg / (1.0 - tau_avg)

    return arm_ratio, filter_order


# noinspection SpellCheckingInspection
def impulse_response_t0_value(
    tau: float, filter_order: int, dt: float, arm_ratio=0.5
) -> float:
    """Returns the h_m(t=0) value.

    The h(t=0) value goes as

                  (1+ar)^2
        h(t=0) =  -------- tau^{-2},   m = 1
                     ar

        h(t=0) = 0, m > 1

    where ar = arm ratio, which is nominally 1/2.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.
    dt: float
        Temporal discretization
    arm_ratio: float
        Ratio of tau_+ / tau_- (default is 1/2).

    Returns
    -------
    float
        Theoretic value of h_m(t=0).
    """

    if filter_order == 1:
        ht0 = np.power(1.0 + arm_ratio, 2) / arm_ratio * np.power(tau, -2)
    else:
        ht0 = 0.0

    return ht0


def zero_crossing_time(tau: float, filter_order: int, arm_ratio=0.5) -> float:
    """The crossing time `tx` such that h(tx) = 0, tx > 0.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        The order of the filter.
    arm_ratio: float
        Ratio of tau_+ / tau_- (default is 1/2).

    Returns
    -------
    float
        Time tx where h_slope(tx) = 0.
    """

    # alias
    tx = (
        2.0
        * arm_ratio
        * np.log(1.0 / arm_ratio)
        / (1.0 - np.power(arm_ratio, 2))
        * filter_order
        / (filter_order + 1.0)
        * tau
    )

    return tx


# noinspection SpellCheckingInspection,PyUnusedLocal
def moment_value_generator(moment: int, tau: float = 1.0, dt: float = 0.0):
    """
    Returns an anonymous function with the single argument (order: int),
    where order is a `filter_order`. The returned function captures the `tau`
    and `dt` arguments in this calling function.

    Parameters
    ----------
    moment: int
        Moment to calculate: [0, 1, 2]
    tau: float
        Temporal scale of the filter.
    dt: float
        Time interval for discretization.

    Returns
    -------
    Anonymous function with argument (order: int).
    """

    # generator defs with `tau` and `dt` capture
    # noinspection PyUnusedLocal
    def m0_gen(order: int) -> float:
        if order == 1:
            return 0.0 + 9.0 / 4.0 * dt / np.power(tau, 2)
        else:
            return 0.0

    # noinspection PyUnusedLocal
    def m1_gen(order: int) -> float:
        return -1.0

    # noinspection PyUnusedLocal
    def m2_gen(order: int) -> float:
        return -2.0 * tau

    # return
    if moment == 0:
        return m0_gen
    elif moment == 1:
        return m1_gen
    elif moment == 2:
        return m2_gen
    else:
        msg = "Moments must be within (0, 1, 2). Calling value: {0}".format(
            moment
        )
        raise IndexError(msg)


# noinspection SpellCheckingInspection
def autocorrelation_peak_and_stride_values(
    tau: float, filter_order: int, dt=0.0
) -> dict:
    """
    System autocorrelation peak and stride values. Peak value is kh(0),
    and the stride comes from the trailing kh(xi_5pct) / kh(0) = -5%.

    Note: `arm_ratio` = 1/2 for these encoded values.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Order of the filter (note: user is required to validate that
        the order is supported).
    dt: float
        [Optional] Temporal increment dt in h(t).

    Returns
    -------
    dict
        Peak value (scaled by 1/tau^3) and 5% pct stride (scaled by tau)
    """

    # build unit_kh0 expr
    m = filter_order
    coef = (
        np.power(9.0, 2 - m)
        * (np.power(9.0, m) - np.power(8.0, m))
        / (32.0 * np.sqrt(np.pi))
    )
    unit_kh0 = (
        coef
        * m
        * scipy.special.gamma(m - 0.5)
        / scipy.special.gamma(m)
        * np.power((m + 1.0) / m, 3)
    )

    # values from Mathematica
    unit_sacf_values = {
        1: {"xi_5pct": 1.91744},
        2: {"xi_5pct": 2.13314},
        3: {"xi_5pct": 2.04505},
        4: {"xi_5pct": 1.94042},
        5: {"xi_5pct": 1.84766},
        6: {"xi_5pct": 1.76889},
        7: {"xi_5pct": 1.70218},
        8: {"xi_5pct": 1.64526},
    }

    # fetch values for order and scale
    sacf_values = unit_sacf_values[filter_order]
    sacf_values["kh0"] = unit_kh0 / np.power(tau, 3)
    sacf_values["kh0"] += autocorrelation_peak_value_discretization_correction(
        tau, filter_order, dt
    )
    sacf_values["xi_5pct"] *= tau

    # return
    return sacf_values


# noinspection SpellCheckingInspection
def autocorrelation_peak_value_discretization_correction(
    tau: float, filter_order: int, dt: float
) -> float:
    """
    Although this is a continuous-time context, the discretization of the
    system autocorrelation function requires a `dt`-order correction. This
    correction can be computed directly by taking the Maclaurin expansion
    of the summed correlogram in terms of dt.

    Parameters
    ----------
    tau: float
        Temporal scale of the filter.
    filter_order: int
        Order of the filter.
    dt: float
        Time-discretization step.

    Returns
    -------
    float
        Discretization correction.
    """

    if filter_order == 1:
        return 81.0 / 8.0 * dt / np.power(tau, 4)
    else:
        return 0.0


# noinspection SpellCheckingInspection,PyPep8Naming
def scale_correlation_stride_and_residual_values(
    tau: float, filter_order: int
) -> dict:
    """Scale decorrelation strides and residual correlations, `arm_ratio` = 1/2.

    Parameters
    ----------
    tau: float
        The first moment of the filter.
    filter_order: int
        Order of the filter.

    Returns
    -------
    dict
        Scale decorrelation (scales as `tau`) and residual correlation
        (independent of `tau`).
    """

    # values from Mathematica
    unit_sscf_values = {
        1: {"sti_5pct": 29.5134, "residual": +0.05},
        2: {"sti_5pct": 44.9399, "residual": -0.05},
        3: {"sti_5pct": 21.3885, "residual": -0.05},
        4: {"sti_5pct": 14.0121, "residual": -0.05},
        5: {"sti_5pct": 10.6693, "residual": -0.05},
        6: {"sti_5pct": 8.80714, "residual": -0.05},
        7: {"sti_5pct": 7.63184, "residual": -0.05},
        8: {"sti_5pct": 6.82619, "residual": -0.05},
    }

    # fetch values for order and scale
    sscf_values = unit_sscf_values[filter_order]
    sscf_values["sti_5pct"] *= tau

    # return
    return sscf_values


# noinspection PyUnusedLocal
def gain_at_dc(tau: float) -> float:
    """Filter gain at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return 0.0


# noinspection PyUnusedLocal
def phase_at_dc(tau: float) -> float:
    """Filter phase at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return np.pi / 2


def group_delay_at_dc(tau: float) -> float:
    """Filter group delay at DC frequency. This is independent of filter order.

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.

    Returns
    -------
    float:
        The gain at DC frequency.
    """

    return tau


# noinspection SpellCheckingInspection
def frequency_of_peak_gain(tau: float, filter_order: int) -> float:
    """Returns the radial frequency at which the gain peaks (arm-ratio=1/2)."""

    # values calculated from Mathematica
    unit_frequency_of_peak_gain = {
        1: 2.12132,
        2: 2.17439,
        3: 2.30366,
        4: 2.43475,
        5: 2.55634,
        6: 2.66689,
        7: 2.76701,
        8: 2.85781,
    }

    return unit_frequency_of_peak_gain[filter_order] / tau


# noinspection SpellCheckingInspection
def gain_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns gain(w_peak), the peak gain (arm-ratio=1/2)."""

    # values calculated from Mathematica
    unit_gain_at_peak_frequency = {
        1: 1.00000,
        2: 1.12577,
        3: 1.23961,
        4: 1.33919,
        5: 1.42674,
        6: 1.50441,
        7: 1.57393,
        8: 1.63663,
    }

    return unit_gain_at_peak_frequency[filter_order] / tau


# noinspection SpellCheckingInspection
def phase_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns phase(w_peak) (arm-ratio=1/2)."""

    # values calculated from Mathematica
    phase_at_peak_frequency_values = {
        1: 0.00000,
        2: -0.23888,
        3: -0.44046,
        4: -0.61179,
        5: -0.75981,
        6: -0.88945,
        7: -1.00429,
        8: -1.10693,
    }

    return phase_at_peak_frequency_values[filter_order]


# noinspection SpellCheckingInspection
def group_delay_at_peak_frequency(tau: float, filter_order: int) -> float:
    """Returns group-delay(w_peak) (arm-ratio=1/2)."""

    # values calculated from Mathematica
    unit_group_delay_at_peak_frequency = {
        1: 0.44444,
        2: 0.60142,
        3: 0.68240,
        4: 0.73221,
        5: 0.76620,
        6: 0.79104,
        7: 0.81009,
        8: 0.82526,
    }

    return unit_group_delay_at_peak_frequency[filter_order] * tau


# noinspection SpellCheckingInspection,PyPep8Naming
def wireframe(tau: float, filter_order: int):
    """Wireframe-timepoint-pair for this slope filter. (arm-ratio = 1/2)

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        The order of the filter.

    Returns
    -------
    WireframeContinuousTime
        Wireframe object
    """

    return design_tools.generate_slope_wireframe_from_spectra(
        sys.modules[__name__], tau, filter_order
    )


# noinspection SpellCheckingInspection
def wavenumber(filter_order: int) -> float:
    """Report the wavenumber for this filter as a function of order.

    Parameters
    ----------
    filter_order: int
        The order of the filter.

    Returns
    -------
    float
        The wavenumber.
    """

    wavenumbers = {
        1: 0.55424844,
        2: 0.68040373,
        3: 0.70865678,
        4: 0.71848684,
        5: 0.72331669,
        6: 0.72664995,
        7: 0.72956863,
        8: 0.73242194,
    }

    return wavenumbers[filter_order]


# noinspection SpellCheckingInspection
def uncertainty_product(filter_order: int) -> dict:
    """Dict of TW, BW, and UCP for this filter order. (arm-ratio = 1/2)

    Parameters
    ----------
    filter_order: int
        The order of the filter

    Returns
    -------
    dict
        TW is temporal width, BW is spectral width, and UCP is the
        uncertainty product.
    """

    uncertainty_products = {
        1: {"TW": 0.40062, "BW": np.inf, "UCP": np.inf},
        2: {"TW": 0.47192, "BW": 5.85204, "UCP": 2.76168},
        3: {"TW": 0.46611, "BW": 4.42917, "UCP": 2.06448},
        4: {"TW": 0.45002, "BW": 4.21715, "UCP": 1.89782},
        5: {"TW": 0.43368, "BW": 4.20664, "UCP": 1.82433},
        6: {"TW": 0.41906, "BW": 4.25805, "UCP": 1.78436},
        7: {"TW": 0.40641, "BW": 4.33153, "UCP": 1.76036},
        8: {"TW": 0.39555, "BW": 4.41225, "UCP": 1.74528},
    }

    return uncertainty_products[filter_order]



"""
-------------------------------------------------------------------------------

Implements a unit step in a discretized continuous time.

-------------------------------------------------------------------------------
"""

import numpy as np

from irides.tools import impulse_response_builder_tools


# noinspection SpellCheckingInspection
def generate_impulse_response(t_end: float, dt: float) -> np.ndarray:
    """Creates a unit-step time series that can be used for convolution.

    Parameters
    ----------
    t_end: float
        End timepoint of the series.
    dt: float
        Time increment.

    Returns
    -------
    np.ndarray
        A time series array (n, 2) having columns [t_axis, h_ustep].
    """

    # init time series
    t_start = 0.0
    (
        time_series,
        i_t0_plus,
    ) = impulse_response_builder_tools.make_impulse_response_template(
        t_start, t_end, dt
    )

    # set the unit step
    time_series[i_t0_plus:, 1] = 1.0

    # return
    return time_series



"""
-------------------------------------------------------------------------------

Discrete-time ideal-delay operator.

-------------------------------------------------------------------------------
"""

import numpy as np
from irides.resources.containers.discrete_sequence import DiscreteSequence


# -----------------------------------------------------------------------------
# design


def delay(mu: float = np.nan) -> float:
    """Returns the average delay of the operator, or nan absent an argument"""

    return mu


# -----------------------------------------------------------------------------
# filter signature


def generate_impulse_response(mu: float) -> DiscreteSequence:
    """Returns a discrete sequence with a single +1 value located at n = mu.

    Parameters
    ----------
    mu: float
        The point of ideal delay. (Note, will be rounded to nearest integer.)

    Returns
    -------
    DiscreteSequence
        A preloaded discrete sequence.
    """

    # round mu (note: mu can take either sign)
    mu_int = np.round(mu).astype(int)

    # determine the length of the n-axis
    n_start = 0
    n_end = 0
    if mu_int >= 0:
        n_end = mu_int
    else:
        n_start = mu_int

    n_end_offset = 1  # cannot exclude a sample for ideal delay

    # setup a discrete sequence
    ds = DiscreteSequence(n_start, n_end + n_end_offset)

    # set the +1 weight
    ds.v_axis = np.zeros_like(ds.n_axis)
    ds.v_axis[ds.index_at(mu_int)] = +1.0

    # return
    return ds



"""
-------------------------------------------------------------------------------

Analog multistage box:

Implements temporal, spectral, autocorrelation and pole / zero features
for this continuous-time level filter.

-------------------------------------------------------------------------------
"""

from typing import Callable
import numpy as np
import sys

from irides.design.analog import mbox_level as design
from irides.tools import impulse_response_builder_tools
from irides.tools import impulse_response_tools
from irides.tools import transfer_function_tools


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_impulse_response(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    """
    Creates a continuous-time-like impulse response for the multistage box
    level filter. The multistage box is constructed from a cascade of
    elementary boxes.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order (number of stages).

    Returns
    -------
    np.ndarray
        A time series array (n, 2) having columns [t_axis, h_mbox].
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # convert
    T = design.convert_tau_to_T(tau)

    # alias
    m = filter_order

    # make time series template
    (
        time_series,
        i_t_0plus,
    ) = impulse_response_builder_tools.make_impulse_response_template(
        t_start, t_end, dt
    )

    # alias
    h = time_series[:, 1]

    # prepare for and exec m-fold convolution
    ref_box = np.ones(int(np.floor(T / dt)))
    conv_box = np.array([1])
    for i in range(m):
        conv_box = np.convolve(ref_box, conv_box)

    # downsample candidate box and normalize gain to unity
    cand_box = conv_box[::m]
    cand_box /= sum(cand_box) * dt

    # make h_mbox by inserting the cand_box into an array of zeros
    i_start = i_t_0plus
    i_end = i_start + cand_box.shape[0]
    h[i_start:i_end] = cand_box

    # return
    return time_series


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_sacf_correlogram(
    xi_start: float,
    xi_end: float,
    dxi: float,
    tau: float,
    filter_order: int,
    normalize=False,
) -> np.ndarray:
    """
    Computes the system autocorrelation response from the correlation
    of the impulse response with itself.

    Parameters
    ----------
    xi_start: float
        Start lag, |xi_start| <= |xi_end|
    xi_end: float
        End lag
    dxi: float
        Lag increment
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order (number of stages).
    normalize: bool
        T/F whether to normalize the sacf spectrum so sacf(0) = 1.

    Returns
    -------
    np.ndarray
        System autocorrelation (n, 2) np.array w/ cols [xi, sacf-value].
    """

    # fetch impulse response
    ir = generate_impulse_response(0.0, xi_end, dxi, tau, filter_order)

    # calc sacf
    sacf = impulse_response_tools.calculate_auto_correlation_function(ir)

    # truncate at xi_start
    sacf = sacf[xi_start - dxi / 2.0 <= sacf[:, 0], :]

    # normalize by theoretic peak value if requested
    if normalize:
        kh0 = design.autocorrelation_peak_and_stride_values(tau, filter_order)[
            "kh0"
        ]
        sacf[:, 1] /= kh0

    return sacf


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_transfer_function(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Computes the transfer function H_mbox(f) of the mutistage box. To avoid a
    numeric overflow for f = 0, the transfer function is expanded in the
    region f ~ 0. We have

                 /
                 |    / 1 - exp(-s T / m) \ m
                 |   |---------------------|             |s| >= 2 sqrt(m) TOL
                 |    \      s T / m      /
        H(s) =  <
                 |
                 |   (1 - s T / 2 + (1+3m)/24m (s T)^2   |s| < 2 sqrt(m) TOL
                 |
                 \

    where s = jw, w = 2pi f, and TOL << 1.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order (number of stages).

    Returns
    -------
    np.ndarray
        The transfer function with cols [f_axis, H_mbox(f)].
    """

    # validate
    impulse_response_tools.validate_filter_order_or_die(design, filter_order)

    # convert
    T = design.convert_tau_to_T(tau)

    # alias
    m = filter_order

    # make transfer-function panel
    npts = np.arange(f_start, f_end, df).shape[0]
    H_mbox = np.ndarray([npts, 2], dtype=complex)

    # setup f_axis
    H_mbox[:, 0] = np.arange(f_start, f_end, df)
    f_axis = H_mbox[:, 0]

    # make f_axis and sTm vector
    sTm = (1j * 2.0 * np.pi * T / m) * f_axis
    sT = m * sTm

    # index into f ~ 0 region and outside that region
    tol = 0.001  # hardcode choice
    thresh = 2.0 * np.sqrt(m) / T * tol
    i_in = np.where(np.abs(f_axis) < thresh)[0]
    i_out = np.where(np.abs(f_axis) >= thresh)[0]

    # calculate transfer function
    H_mbox[i_in, 1] = (
        1.0
        - 0.5 * sT[i_in]
        + (1.0 + 3.0 * m) / (24.0 * m) * np.power(sT[i_in], 2)
    )

    H_mbox[i_out, 1] = np.power((1.0 - np.exp(-sTm[i_out])) / sTm[i_out], m)

    # return
    return H_mbox


# noinspection SpellCheckingInspection,PyPep8Naming
def make_callable_transfer_function(
    filter_order: int,
) -> Callable[[float, float], complex]:
    """Returns a callable function with signature (f, tau) -> complex.

    This transfer function is restricted to the jw axis:

        H(s) -> H(j w).

    Parameters
    ----------
    filter_order: int
        The filter order, which is captured by the definition herein.

    Returns
    -------
    Callable[[float, float], complex]
        A callable function (f: float, tau: float) -> complex
    """

    # noinspection PyPep8Naming
    def H(f: float, tau: float = 1.0) -> complex:
        """Compute complex xfer fcxn H at freq `f` with optional scale tau.

        Note that `filter_order` is captured, leaving only (f, tau) as free
        variables.

        Parameters
        ----------
        f: float
            Start cyclic frequency.
        tau: float
            Temporal scaling, which ='s the first moment of the final filter.

        Returns
        -------
        complex
            H(f, tau) value.
        """

        jw = 1j * 2.0 * np.pi * f
        T = 2.0 * tau
        sTm = jw * T / filter_order

        return np.power((1.0 - np.exp(-sTm)) / sTm, filter_order)

    return H


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_sscf_correlogram(
    sti_start: float, sti_end: float, dsti: float, filter_order: int
) -> np.ndarray:
    """
    Computes the normalized system scale-correlation correlogram from the
    filter's transfer function.

    No sti-normalization correction is necessary for level filters.

    Parameters
    ----------
    sti_start: float
        Start scale, |sti_start| <= |sti_end|
    sti_end: float
        End scale
    dsti: float
        Scale increment
    filter_order: int
        Filter order.

    Returns
    -------
    np.ndarray
        System scale-correlation normalized (n, 2)
        np.array w/ cols [sti, sscf-value].
    """

    # correction function
    sti_norm_correction_fcxn = lambda sti: 1.0

    # return sscf-norm np-array
    return transfer_function_tools.calculate_scale_correlation_function(
        sti_start,
        sti_end,
        dsti,
        filter_order,
        sys.modules[__name__],
        sti_norm_correction_fcxn,
    )


# noinspection SpellCheckingInspection,PyPep8Naming,PyUnusedLocal
def generate_phase_spectrum(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    r"""
    Computes the phase of H_mbox(f) from its analytic expression. The phase
    spectrum is independent of filter_order (in this case) and is simply

        phase-H_mbox(f) = - (2 pi f) T / 2.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order (number of stages).  (unused)

    Returns
    -------
    np.ndarray
        The phase spectrum with cols [f_axis, AngleH_mbox(f)].
    """

    # convert
    T = design.convert_tau_to_T(tau)

    # make phase-spectrum panel
    npts = np.arange(f_start, f_end, df).shape[0]
    angle_H_mbox = np.ndarray([npts, 2], dtype=float)

    # setup f_axis
    angle_H_mbox[:, 0] = np.arange(f_start, f_end, df)
    f_axis = angle_H_mbox[:, 0]

    # compute the phase
    angle_H_mbox[:, 1] = -np.pi * T * f_axis

    return angle_H_mbox


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_spectra(
    f_start: float, f_end: float, df: float, tau: float, filter_order: int
) -> np.ndarray:
    """
    Generate gain, phase and group-delay spectra, based on the transfer
    function.

    Parameters
    ----------
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.
    df: float
        Cyclic frequency increment.
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order (number of stages).

    Returns
    -------
    np.ndarray
        Spectra panel [n_pts, 4] with cols [f, g(f), phase(f), grpdly(f)].
    """

    # fetch the transfer function for gain calculation
    H: np.ndarray = generate_transfer_function(
        f_start, f_end, df, tau, filter_order
    )

    # fetch the phase spectrum for group-delay calculation
    phase_H: np.ndarray = generate_phase_spectrum(
        f_start, f_end, df, tau, filter_order
    )

    # calculate and return spectra
    return transfer_function_tools.calculate_spectra(H, phase_H)


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_wireframe(
    t_start: float, t_end: float, dt: float, tau: float, filter_order: int
) -> np.ndarray:
    """Returns a wireframe equivalent for this filter.

    Parameters
    ----------
    t_start: float
        Start time.
    t_end: float
        End time.
    dt: float
        Time increment.
    tau: float
        First moment of the filter.
    filter_order: int
        Filter order.

    Returns
    -------
    WireframeContinuousTime
        A wireframe container.
    """

    # fetch the wireframe, create a timeseries template
    wf = design.wireframe(tau, filter_order)
    ts = impulse_response_builder_tools.discretize_wireframe_onto_time_grid(
        t_start, t_end, dt, wf
    )

    return ts


# noinspection SpellCheckingInspection,PyPep8Naming
def generate_poles_and_zeros(
    tau: float, filter_order: int, f_start=0.0, f_end=0.0
) -> dict:
    """
    Calculates poles and zeros for the multistage box in the s-plane.
    There are no poles, and zeros are periodic along the jw axis, with
    the exception of the origin (w, sigma) = 0, where there is no zero.

    The zero positions are

        f[n] = n / (T / m),  n = +/- 1, +/- 2, ...  (n != 0).

    Parameters
    ----------
    tau: float
        Temporal scaling, which ='s the first moment of the final filter.
    filter_order: int
        Filter order (number of stages).
    f_start: float
        Start cyclic frequency.
    f_end: float
        End cyclic frequency.

    Returns
    -------
    dict
        Dictionary with keys ('poles', 'zeros') and values as np.ndarray
        with dtype=complex.
    """

    # convert
    T = design.convert_tau_to_T(tau)

    # the request is for domain [f_min, f_max], and this needs to be
    # converted to the domain [n_min, n_max]. The trick is to pull toward
    # zero frequency for either frequency sign. That is,
    # np.floor(2.2) = 2.0 while np.floor(-2.2) = -3.0.
    n_start = np.sign(f_start) * np.floor(
        np.floor(np.abs(f_start)) * (T / filter_order)
    )
    n_end = np.sign(f_end) * np.floor(
        np.floor(np.abs(f_end)) * (T / filter_order)
    )

    n_axis = np.arange(n_start, n_end + 1, dtype=int)
    n_axis = np.delete(n_axis, np.where(n_axis == 0)[0])  # remove any n=0 entry

    # zero locations
    zeros = np.zeros(n_axis.shape[0], dtype=complex)
    zeros[:] = np.zeros_like(n_axis) + 1j * n_axis / (T / filter_order)

    # fill in poles_zeros dict
    poles_zeros = {"poles": np.ndarray((0, 0), dtype=complex), "zeros": zeros}

    # return
    return poles_zeros
